import base64
import importlib.metadata
import json
import os
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import urlsplit

import click
import yaml
from pydantic import BaseModel, Field

from ..cityagent import SocietyAgent

version_string_of_agentsociety = importlib.metadata.version("agentsociety")

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
PARAM_DECLS_VERSION = ("-V", "--version")


def load_config(
    config_path: Optional[str] = None, config_base64: Optional[str] = None
) -> Dict[Any, Any]:
    """Load configuration file, supports JSON and YAML formats, as well as base64 encoded config"""
    if config_base64:
        # Decode base64 config
        config_str = base64.b64decode(config_base64).decode("utf-8")
        # For base64 encoded config, try both JSON and YAML parsing
        try:
            # Try to parse as JSON
            return json.loads(config_str)
        except json.JSONDecodeError:
            try:
                # Try to parse as YAML
                return yaml.safe_load(config_str)
            except yaml.YAMLError as e:
                raise click.BadParameter(f"Failed to parse base64 config: {e}")
    elif config_path:
        # Determine format based on file extension
        path = Path(config_path)
        if not path.exists():
            raise click.BadParameter(f"Config file {config_path} does not exist")
        file_ext = path.suffix.lower()
        if file_ext in [".json"]:
            try:
                with open(config_path, "r", encoding="utf-8") as f:
                    return json.load(f)
            except json.JSONDecodeError as e:
                raise click.BadParameter(f"Failed to parse JSON config file: {e}")
        elif file_ext in [".yaml", ".yml"]:
            try:
                with open(config_path, "r", encoding="utf-8") as f:
                    return yaml.safe_load(f)
            except yaml.YAMLError as e:
                raise click.BadParameter(f"Failed to parse YAML config file: {e}")
        else:
            raise click.BadParameter(f"Unsupported config file format: {file_ext}")
    else:
        raise click.BadParameter("No config file or base64 encoded config provided")


def common_options(f):
    """Common configuration options decorator"""
    f = click.option("--config", "-c", help="Path to config file (.json, .yml, .yaml)")(
        f
    )
    f = click.option(
        "--config-base64", help="Base64 encoded config with JSON or YAML format"
    )(f)
    return f


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version_string_of_agentsociety, prog_name="AgentSociety")
def cli():
    """AgentSociety CLI tool"""
    pass


@cli.command()
@common_options
def ui(config: Optional[str], config_base64: Optional[str]):
    """Launch AgentSociety GUI"""
    if config is None and config_base64 is None:
        config_data = {}
    else:
        config_data = load_config(config, config_base64)

    import uvicorn

    from ..configs import EnvConfig
    from ..logger import get_logger, set_logger_level
    from ..webapi.app import create_app
    from ..executor import ProcessExecutor

    class WebUIConfig(BaseModel):
        addr: str = Field(default="127.0.0.1:8080")
        callback_url: str = ""
        env: EnvConfig = Field(default_factory=lambda: EnvConfig.model_validate({
            "db": {
                "enabled": True,
            }
        }))
        read_only: bool = Field(default=False)
        debug: bool = Field(default=False)
        logging_level: str = Field(default="INFO")
        commercial: Dict[str, Any] = {}

    async def _main():
        c = WebUIConfig.model_validate(config_data)
        set_logger_level(c.logging_level.upper())
        get_logger().info("Launching AgentSociety WebUI")
        get_logger().debug(f"WebUI config: {c}")

        db_dsn = c.env.db.get_dsn(Path(c.env.home_dir) / "sqlite.db")

        # default executor
        executor = ProcessExecutor(c.env.home_dir)

        more_state: Dict[str, Any] = {
            "callback_url": c.callback_url,
            "executor": executor,
        }

        app = create_app(
            db_dsn=db_dsn,
            read_only=c.read_only,
            env=c.env,
            more_state=more_state,
            commercial=c.commercial,
        )

        # Start server
        url = urlsplit("//" + c.addr)
        host, port = url.hostname, url.port
        if host is None or host == "" or host == "localhost":
            host = "127.0.0.1"
        if port is None:
            port = 8080
        log_level = "debug" if c.debug else "info"
        get_logger().info("Starting server at %s:%s", host, port)
        config = uvicorn.Config(app, host=host, port=port, log_level=log_level)
        server = uvicorn.Server(config)
        await server.serve()

    import asyncio

    asyncio.run(_main())


@cli.command()
@common_options
@click.option("--type", default="simulation", help="Specify the type of the simulation", type=click.Choice(["simulation", "individual"]))
def check(config: str, config_base64: str, type: str):
    """Pre-check the config"""
    from pathlib import Path

    config_dict = load_config(config, config_base64)

    if type == "simulation":
        from ..configs import Config

        c = Config.model_validate(config_dict)
    elif type == "individual":
        from ..configs import IndividualConfig

        c = IndividualConfig.model_validate(config_dict)
    else:
        raise ValueError(f"Invalid type: {type}")

    click.echo(f"Config format check. {click.style('Passed.', fg='green')}")

    # =================
    # check the connection to the database server using SQLAlchemy
    # =================
    if c.env.db.enabled:
        from sqlalchemy import create_engine, text
        from sqlalchemy.exc import OperationalError

        try:
            # Get the DSN for the database
            sqlite_path = Path(c.env.home_dir) / "sqlite.db"
            dsn = c.env.db.get_dsn(sqlite_path)

            # Convert async DSN to sync DSN for testing
            if dsn.startswith("postgresql+asyncpg://"):
                sync_dsn = dsn.replace("postgresql+asyncpg://", "postgresql://", 1)
            elif dsn.startswith("sqlite+aiosqlite://"):
                sync_dsn = dsn.replace("sqlite+aiosqlite://", "sqlite://", 1)
            else:
                sync_dsn = dsn

            # Create sync engine for testing
            engine = create_engine(sync_dsn)

            # Test connection
            with engine.connect() as conn:
                conn.execute(text("SELECT 1"))

            click.echo(f"Database connection check. {click.style('Passed.', fg='green')}")

        except OperationalError as e:
            click.echo(f"Database connection check. {click.style('Failed:', fg='red')} {e}")
            if c.env.db.db_type == "postgresql":
                click.echo(
                    f"Explanation: Please check the `pg_dsn` (value={c.env.db.pg_dsn}) of the PostgreSQL server. The format of `pg_dsn` is `postgresql://<username>:<password>@<host>:<port>/<database_name>`. The item wrapped in `<>` should be replaced with the actual values."
                )
                error_msg = str(e)
                if "password authentication failed" in error_msg:
                    click.echo(
                        "Explanation: The username or password of the PostgreSQL server is incorrect."
                    )
                elif "Temporary failure in name resolution" in error_msg:
                    click.echo(
                        "Explanation: The host of the PostgreSQL server is invalid. Maybe you should use `localhost` or `127.0.0.1` instead if you are running the simulation on a single machine (used to run docker compose)."
                    )
                elif "Connection refused" in error_msg:
                    click.echo(
                        "Explanation: The host or port of the PostgreSQL server is incorrect."
                    )
            elif c.env.db.db_type == "sqlite":
                click.echo(
                    "Explanation: SQLite database connection failed. Please check the file path and permissions."
                )
        except Exception as e:
            click.echo(f"Database connection check. {click.style('Failed:', fg='red')} {e}")
    else:
        click.echo(f"Database is disabled. {click.style('Skipped.', fg='yellow')}")

    # =================
    # check whether the map file exists
    # =================
    if type == "simulation":
        if not os.path.exists(c.map.file_path): # type: ignore
            click.echo(
                f"Map file {c.map.file_path} does not exist. {click.style('Failed.', fg='red')}" # type: ignore
            )
            return
        click.echo(f"Map file. {click.style('Passed.', fg='green')}")


@cli.command()
@common_options
@click.option("--tenant-id", default="default", help="Specify tenant ID")
@click.option("--callback-url", default="", help="Specify callback URL (POST)")
@click.option("--type", default="simulation", help="Specify the type of the simulation", type=click.Choice(["simulation", "individual"]))
def run(
    config: str,
    config_base64: str,
    tenant_id: str,
    callback_url: str,
    type: str,
):
    """Run the simulation"""
    config_dict = load_config(config, config_base64)

    import requests
    from ..configs import Config, IndividualConfig
    from ..simulation import AgentSociety
    from ..cityagent import default
    from ..configs.exp import AgentFilterConfig

    if type == "simulation":
        c = Config.model_validate(config_dict)
        # Check if we need to import community modules
        need_community = False
        if c.agents.citizens:
            for citizen in c.agents.citizens:
                if isinstance(citizen.agent_class, str):
                    need_community = True
                    break
                # go to check blocks
                if citizen.blocks is not None:
                    for block in citizen.blocks.keys():
                        if isinstance(block, str):
                            need_community = True
                            break

        if c.agents.supervisor is not None and isinstance(
            c.agents.supervisor.agent_class, str
        ):
            need_community = True

        if c.exp.workflow:
            for step in c.exp.workflow:
                if step.func is not None and isinstance(step.func, str):
                    need_community = True
                    break
                # Check if target_agent is AgentFilterConfig with agent_class
                if (step.target_agent is not None and 
                    isinstance(step.target_agent, AgentFilterConfig) and 
                    step.target_agent.agent_class is not None):
                    need_community = True
                    break

        if need_community:
            try:
                from agentsociety_community.agents import citizens, supervisors
                from agentsociety_community.blocks import citizens as citizen_blocks
                from agentsociety_community.workflows import functions as workflow_functions

                # get the mapping of the agent class
                workflow_function_map = workflow_functions.get_type_to_cls_dict()
                citizens_class_map = citizens.get_type_to_cls_dict()
                citizen_blocks_class_map = citizen_blocks.get_type_to_cls_dict()
                supervisors_class_map = supervisors.get_type_to_cls_dict()

                # process the agent_class in citizens
                for citizen in c.agents.citizens:
                    if isinstance(citizen.agent_class, str):
                        if citizen.agent_class == "SocietyAgent" or citizen.agent_class == "citizen":
                            citizen.agent_class = 'citizen'
                        else:
                            citizen.agent_class = citizens_class_map[citizen.agent_class]()
                    if citizen.blocks is not None:
                        new_blocks = {}
                        for block_name, block_params in citizen.blocks.items():
                            if isinstance(block_name, str):
                                block_class = citizen_blocks_class_map[block_name]()
                                new_blocks[block_class] = block_params
                            else:
                                new_blocks[block_name] = block_params
                        citizen.blocks = new_blocks

                # process the agent_class in supervisor
                if c.agents.supervisor is not None and isinstance(
                    c.agents.supervisor.agent_class, str
                ):
                    c.agents.supervisor.agent_class = supervisors_class_map[
                        c.agents.supervisor.agent_class
                    ]()

                # process the func in workflow
                for step in c.exp.workflow:
                    if step.func is not None and isinstance(step.func, str):
                        # if func is a string, try to get the corresponding function from function_map
                        imported_func = workflow_function_map[step.func]()
                        step.func = imported_func
                    
                    # process the agent_class in target_agent if it's an AgentFilterConfig
                    if step.target_agent is not None and isinstance(step.target_agent, AgentFilterConfig):
                        agent_filter = step.target_agent
                        if agent_filter.agent_class is not None:
                            if isinstance(agent_filter.agent_class, list):
                                # Convert string list to class list
                                converted_classes = []
                                for agent_class_str in agent_filter.agent_class:
                                    if isinstance(agent_class_str, str):
                                        if agent_class_str == "citizen":
                                            converted_classes.append(SocietyAgent)
                                        elif agent_class_str == "SocietyAgent":
                                            converted_classes.append(SocietyAgent)
                                        elif agent_class_str in citizens_class_map:
                                            converted_classes.append(citizens_class_map[agent_class_str]())
                                        elif agent_class_str in supervisors_class_map:
                                            converted_classes.append(supervisors_class_map[agent_class_str]())
                                        else:
                                            # Keep as string if not found in maps
                                            converted_classes.append(agent_class_str)
                                    else:
                                        # Already a class, keep as is
                                        converted_classes.append(agent_class_str)
                                agent_filter.agent_class = tuple(converted_classes)  # type: ignore
                            elif isinstance(agent_filter.agent_class, str):
                                # Convert single string to class
                                if agent_filter.agent_class == "citizen":
                                    agent_filter.agent_class = tuple(SocietyAgent,) # type: ignore
                                elif agent_filter.agent_class == "SocietyAgent":
                                    # For default config compatible
                                    agent_filter.agent_class = tuple(SocietyAgent,) # type: ignore
                                if agent_filter.agent_class in citizens_class_map:
                                    agent_filter.agent_class = tuple(citizens_class_map[agent_filter.agent_class](),)  # type: ignore
                                elif agent_filter.agent_class in supervisors_class_map:
                                    agent_filter.agent_class = tuple(supervisors_class_map[agent_filter.agent_class](),)  # type: ignore
                                # If not found in maps, keep as string
            except ImportError as e:
                import traceback
                print(traceback.format_exc())
                print(
                    "agentsociety_community is not installed. Please install it with `pip install agentsociety-community`"
                )
                raise e
            
        c = default(c)
    elif type == "individual":
        c = IndividualConfig.model_validate(config_dict)
        # Check if we need to import community modules
        need_community = False
        if c.individual.agent_class is not None:
            if isinstance(c.individual.agent_class, str):
                need_community = True
            # go to check blocks
            if c.individual.blocks is not None:
                for block in c.individual.blocks.keys():
                    if isinstance(block, str):
                        need_community = True
                        break
        if need_community:
            try:
                from agentsociety_community.agents import citizens, supervisors
                from agentsociety_community.blocks import citizens as citizen_blocks
                from agentsociety_community.workflows import functions as workflow_functions

                # get the mapping of the agent class
                workflow_function_map = workflow_functions.get_type_to_cls_dict()
                citizens_class_map = citizens.get_type_to_cls_dict()
                citizen_blocks_class_map = citizen_blocks.get_type_to_cls_dict()
                supervisors_class_map = supervisors.get_type_to_cls_dict()

                # process the agent_class in solver
                if c.individual.agent_class is not None and isinstance(c.individual.agent_class, str):
                    c.individual.agent_class = citizens_class_map[c.individual.agent_class]()

                # process the blocks in solver
                if c.individual.blocks is not None:
                    new_blocks = {}
                    for block_name, block_params in c.individual.blocks.items():
                        if isinstance(block_name, str):
                            block_class = citizen_blocks_class_map[block_name]()
                            new_blocks[block_class] = block_params
                        else:
                            new_blocks[block_name] = block_params
                    c.individual.blocks = new_blocks
            except ImportError as e:
                import traceback

                print(traceback.format_exc())
                print(
                    "agentsociety_community is not installed. Please install it with `pip install agentsociety-community`"
                )
                raise e
    else:
        raise ValueError(f"Invalid type: {type}")

    society = AgentSociety.create(c, tenant_id)

    async def _run():
        try:
            await society.init()
            await society.run()
        finally:
            await society.close()
            if callback_url:
                requests.post(callback_url)

    import asyncio

    asyncio.run(_run())


if __name__ == "__main__":
    cli()



import random
from typing import Any, Optional

import json_repair
from pydantic import Field

from ...agent import (
    AgentToolbox,
    Block,
    FormatPrompt,
    BlockParams,
    BlockContext,
    DotDict,
)
from ...logger import get_logger
from ...memory import Memory
from ...agent.dispatcher import BlockDispatcher
from .utils import TIME_ESTIMATE_PROMPT, clean_json_response
from ..sharing_params import SocietyAgentBlockOutput


SLEEP_TIME_ESTIMATION_PROMPT = """As an intelligent agent's time estimation system, please estimate the time needed to complete the current action based on the overall plan and current intention.

Overall plan:
${context.plan_context["plan"]}

Current action: ${context.current_step["intention"]}

Current emotion: ${status.emotion_types}

Examples:
- "Learn programming": {{"time": 120}}
- "Watch a movie": {{"time": 150}} 
- "Play mobile games": {{"time": 60}}
- "Read a book": {{"time": 90}}
- "Exercise": {{"time": 45}}

Please return the result in JSON format (Do not return any other text), the time unit is [minute], example:
{{
    "time": 10
}}
"""


class SleepBlock(Block):
    """Block implementation for handling sleep-related actions in an agent's workflow.

    Attributes:
        description (str): Human-readable block purpose.
        guidance_prompt (FormatPrompt): Template for generating time estimation prompts.
    """

    name = "SleepBlock"
    description = "Handles sleep-related actions"

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Optional[Memory] = None,
        sleep_time_estimation_prompt: str = SLEEP_TIME_ESTIMATION_PROMPT,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.guidance_prompt = FormatPrompt(
            template=sleep_time_estimation_prompt,
            memory=agent_memory,
        )

    async def forward(self, context: DotDict):
        """Execute sleep action and estimate time consumption using LLM.

        Args:
            context: Workflow context containing plan and other metadata.

        Returns:
            Dictionary with execution status, evaluation, time consumed, and node ID.
        """
        await self.guidance_prompt.format(context=context)
        result = await self.llm.atext_request(
            self.guidance_prompt.to_dialog(), response_format={"type": "json_object"}
        )
        result = clean_json_response(result)
        node_id = await self.memory.stream.add(
            topic="other", description="I slept"
        )
        try:
            result: Any = json_repair.loads(result)
            return {
                "success": True,
                "evaluation": f'Sleep: {context["current_step"]["intention"]}',
                "consumed_time": result["time"],
                "node_id": node_id,
            }
        except Exception as e:
            get_logger().warning(
                f"An error occurred while evaluating the response at parse time: {str(e)}, original result: {result}"
            )
            return {
                "success": True,
                "evaluation": f'Sleep: {context["current_step"]["intention"]}',
                "consumed_time": random.randint(1, 8) * 60,
                "node_id": node_id,
            }


class OtherNoneBlock(Block):
    """Fallback block for handling undefined/non-specific actions in workflows.

    Attributes:
        description (str): Human-readable block purpose.
        guidance_prompt (FormatPrompt): Template for generating time estimation prompts.
    """

    name = "OtherNoneBlock"
    description = "Handles all kinds of intentions/actions except sleep"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Optional[Memory] = None):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.guidance_prompt = FormatPrompt(template=TIME_ESTIMATE_PROMPT)

    async def forward(self, context: DotDict):
        await self.guidance_prompt.format(
            plan=context["plan_context"]["plan"],
            intention=context["current_step"]["intention"],
            emotion_types=await self.memory.status.get("emotion_types"),
        )
        result = await self.llm.atext_request(
            self.guidance_prompt.to_dialog(), response_format={"type": "json_object"}
        )
        result = clean_json_response(result)
        node_id = await self.memory.stream.add(
            topic="other",
            description=f"I {context['current_step']['intention']}"
        )
        try:
            result: Any = json_repair.loads(result)
            return {
                "success": True,
                "evaluation": f'Finished executing {context["current_step"]["intention"]}',
                "consumed_time": result["time"],
                "node_id": node_id,
            }
        except Exception as e:
            get_logger().warning(
                f"An error occurred while evaluating the response at parse time: {str(e)}, original result: {result}"
            )
            return {
                "success": True,
                "evaluation": f'Finished executing {context["current_step"]["intention"]}',
                "consumed_time": random.randint(1, 180),
                "node_id": node_id,
            }


class OtherBlockParams(BlockParams):
    sleep_time_estimation_prompt: str = Field(
        default=SLEEP_TIME_ESTIMATION_PROMPT,
        description="Used to determine the sleep time",
    )


class OtherBlockContext(BlockContext): ...


class OtherBlock(Block):
    """Orchestration block for managing specialized sub-blocks (SleepBlock/OtherNoneBlock).

    Attributes:
        sleep_block (SleepBlock): Specialized block for sleep actions.
        other_none_block (OtherNoneBlock): Fallback block for generic actions.
        trigger_time (int): Counter for block activation frequency.
        token_consumption (int): Accumulated LLM token usage.
        dispatcher (BlockDispatcher): Router for selecting appropriate sub-blocks.
    """

    ParamsType = OtherBlockParams
    OutputType = SocietyAgentBlockOutput
    ContextType = OtherBlockContext
    name = "OtherBlock"
    description = "Responsible for all kinds of intentions/actions except mobility, economy, and social, for example, sleep, other actions, etc."
    actions = {
        "sleep": "Support the sleep action",
        "other": "Support other actions",
    }

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        block_params: Optional[OtherBlockParams] = None,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
            block_params=block_params,
        )
        # init all blocks
        self.sleep_block = SleepBlock(
            toolbox, agent_memory, self.params.sleep_time_estimation_prompt
        )
        self.other_none_block = OtherNoneBlock(toolbox, agent_memory)
        self.trigger_time = 0
        self.token_consumption = 0
        # init dispatcher
        self.dispatcher = BlockDispatcher(toolbox, agent_memory)
        # register all blocks
        self.dispatcher.register_blocks([self.sleep_block, self.other_none_block])

    async def forward(self, agent_context: DotDict) -> SocietyAgentBlockOutput:
        """Route workflow steps to appropriate sub-blocks and track resource usage.

        Args:
            context: Workflow context containing plan and metadata.

        Returns:
            Execution result from the selected sub-block.
        """
        self.trigger_time += 1
        consumption_start = (
            self.llm.prompt_tokens_used + self.llm.completion_tokens_used
        )

        context = agent_context | self.context

        # Select the appropriate sub-block using dispatcher
        selected_block = await self.dispatcher.dispatch(context)

        if selected_block is None:
            node_id = await self.memory.stream.add(
                topic="other",
                description=f"I {context['current_step']['intention']}"
            )
            return self.OutputType(
                success=True,
                evaluation=f"Successfully {context['current_step']['intention']}",
                consumed_time=random.randint(1, 30),
                node_id=node_id,
            )

        # Execute the selected sub-block and get the result
        result = await selected_block.forward(context)

        consumption_end = self.llm.prompt_tokens_used + self.llm.completion_tokens_used
        self.token_consumption += consumption_end - consumption_start

        return self.OutputType(**result)



from typing import Optional

from pydantic import Field
from ..agent import (
    BlockOutput,
    AgentParams,
    AgentContext,
)
from ..agent.dispatcher import DISPATCHER_PROMPT
from .blocks.needs_block import INITIAL_NEEDS_PROMPT
from .blocks.plan_block import DETAILED_PLAN_PROMPT


class SocietyAgentConfig(AgentParams):
    """Configuration for society agent."""
    block_dispatch_prompt: str = Field(
        default=DISPATCHER_PROMPT,
        description="The prompt used for the block dispatcher, there is a variable 'intention' in the prompt, which is the intention of the task, used to select the most appropriate block",
    )

    enable_cognition: bool = Field(
        default=True, description="Whether to enable cognition"
    )

    # Maxlow's Needs
    need_initialization_prompt: str = Field(
        default=INITIAL_NEEDS_PROMPT, description="Initial needs prompt"
    )

    # Planned-Behavior
    max_plan_steps: int = Field(
        default=6, description="Maximum number of steps in a plan"
    )
    plan_generation_prompt: str = Field(
        default=DETAILED_PLAN_PROMPT, description="Plan generation prompt"
    )


class SocietyAgentBlockOutput(BlockOutput):
    success: bool  # whether the action is executed successfully
    evaluation: str  # evaluation of the action
    consumed_time: int  # time consumed by the action
    node_id: Optional[int]  # stream memory node id


class SocietyAgentContext(AgentContext):
    # Normal Information
    current_time: str = Field(
        default="Don't know", description="Current time, in HH:MM:SS format"
    )
    current_need: str = Field(
        default="Don't know", description="Current need"
    )  # runtime
    current_intention: str = Field(
        default="Don't know", description="Current intention"
    )  # runtime
    current_emotion: str = Field(default="Don't know", description="Current emotion")
    current_thought: str = Field(default="Don't know", description="Current thought")
    current_location: str = Field(default="Don't know", description="Current location")

    # Environment Information
    area_information: str = Field(default="Don't know", description="Area information")
    weather: str = Field(default="Don't know", description="Weather information")
    temperature: str = Field(
        default="Don't know", description="Temperature information"
    )
    other_information: str = Field(
        default="Don't know", description="Other information"
    )

    # Plan Information
    plan_target: str = Field(default="Don't know", description="Plan target")
    max_plan_steps: int = Field(
        default=6, description="Maximum number of steps in a plan"
    )

    # Block Execution Information
    current_step: dict = Field(default={}, description="Current step")
    plan_context: dict = Field(default={}, description="Plan context")



import numbers
import random
from typing import Any, Optional

import json_repair
import numpy as np
from pydantic import Field

from ...agent import (
    AgentToolbox,
    Block,
    FormatPrompt,
    BlockParams,
    BlockDispatcher,
    DotDict,
    BlockContext,
)
from ...logger import get_logger
from ...memory import Memory
from ..sharing_params import SocietyAgentBlockOutput
from .utils import clean_json_response, prettify_document
from .utils import extract_dict_from_string

WORKTIME_ESTIMATE_PROMPT = """As an intelligent agent's time estimation system, please estimate the time needed to complete the current action based on the overall plan and current intention.

Overall plan:
${context.plan_context["plan"]}

Current action: ${context.current_step["intention"]}

Current emotion: ${status.emotion_types}

Examples:
- "Learn programming": {{"time": 120}}
- "Watch a movie": {{"time": 150}} 
- "Play mobile games": {{"time": 60}}
- "Read a book": {{"time": 90}}
- "Exercise": {{"time": 45}}

Please return the result in JSON format (Do not return any other text), the time unit is [minute], example:
{{
    "time": 10
}}
"""


def softmax(x, gamma=1.0):
    """Compute softmax values with temperature scaling.

    Args:
        x: Input values (array or list)
        gamma: Temperature parameter (higher values make distribution sharper)

    Returns:
        Probability distribution over input values
    """
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    x *= gamma
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=-1, keepdims=True)


class WorkBlock(Block):
    """Handles work-related economic activities and time tracking.

    Attributes:
        guidance_prompt: Template for time estimation queries
    """

    name = "WorkBlock"
    description = "Handles work-related economic activities and time tracking"

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        worktime_estimation_prompt: str = WORKTIME_ESTIMATE_PROMPT,
    ):
        """Initialize with dependencies.

        Args:
            llm: Language model for time estimation
            environment: Time tracking environment
            memory: Agent's memory system
        """
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.guidance_prompt = FormatPrompt(
            template=worktime_estimation_prompt,
            memory=agent_memory,
        )

    async def forward(self, context: DotDict):
        """Process work task and track time expenditure.

        Workflow:
            1. Format prompt with work context
            2. Request time estimation from LLM
            3. Record work experience in memory
            4. Fallback to random time on parsing failures

        Returns:
            Execution result with time consumption details
        """
        await self.guidance_prompt.format(context=context)
        result = await self.llm.atext_request(
            self.guidance_prompt.to_dialog(), response_format={"type": "json_object"}
        )
        result = clean_json_response(result)
        try:
            result: Any = json_repair.loads(result)
            time = result["time"]
            day, start_time = self.environment.get_datetime(format_time=True)
            await self.memory.status.update(
                "working_experience",
                [
                    f"Start from {start_time}, worked {time} minutes on {context['current_step']['intention']}"
                ],
                mode="merge",
            )
            work_hour_finish = await self.memory.status.get("work_hour_finish")
            work_hour_finish += float(time / 60)
            node_id = await self.memory.stream.add(
                topic="economy",
                description=f"I worked {time} minutes on {context['current_step']['intention']}",
            )
            await self.memory.status.update("work_hour_finish", work_hour_finish)
            return {
                "success": True,
                "evaluation": f'work: {context["current_step"]["intention"]}',
                "consumed_time": time,
                "node_id": node_id,
            }
        except Exception as e:
            get_logger().warning(f"Error in parsing: {str(e)}, raw: {result}")
            time = random.randint(1, 3) * 60
            day, start_time = self.environment.get_datetime(format_time=True)
            await self.memory.status.update(
                "working_experience",
                [
                    f"Start from {start_time}, worked {time} minutes on {context['current_step']['intention']}"
                ],
                mode="merge",
            )
            work_hour_finish = await self.memory.status.get("work_hour_finish")
            node_id = await self.memory.stream.add(
                topic="economy",
                description=f"I worked {time} minutes on {context['current_step']['intention']}",
            )
            work_hour_finish += float(time / 60)
            await self.memory.status.update("work_hour_finish", work_hour_finish)
            return {
                "success": True,
                "evaluation": f'work: {context["current_step"]["intention"]}',
                "consumed_time": time,
                "node_id": node_id,
            }


class ConsumptionBlock(Block):
    """Manages consumption behavior and budget constraints.

    Attributes:
        economy_client: Interface to economic simulation
        forward_times: Counter for execution attempts
    """

    name = "ConsumptionBlock"
    description = "Used to determine the consumption amount, and items"

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
    ):
        """Initialize consumption processor.

        Args:
            economy_client: Client for economic system interactions
        """
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.forward_times = 0

    async def forward(self, context: DotDict):
        """Execute consumption decision-making.

        Workflow:
            1. Check monthly consumption limits
            2. Calculate price-weighted demand distribution
            3. Execute transactions through economy client
            4. Record consumption in memory stream

        Returns:
            Consumption evaluation with financial details
        """
        self.forward_times += 1
        agent_id = await self.memory.status.get("id")  # agent_id
        firms_id = await self.environment.economy_client.get_firm_ids()
        intention = context["current_step"]["intention"]
        month_consumption = await self.memory.status.get("to_consumption_currency")
        consumption_currency = await self.environment.economy_client.get(
            agent_id, "consumption"
        )
        if consumption_currency >= month_consumption:
            node_id = await self.memory.stream.add(
                topic="economy",
                description="I have passed the monthly consumption limit, so I will not consume.",
            )
            return {
                "success": False,
                "evaluation": "I have passed the monthly consumption limit, so I will not consume.",
                "consumed_time": 0,
                "node_id": node_id,
            }
        consumption = min(
            month_consumption / 1, month_consumption - consumption_currency
        )
        prices = await self.environment.economy_client.get(firms_id, "price")
        consumption_each_firm = consumption * softmax(prices, gamma=-0.01)
        demand_each_firm = []
        for i in range(len(firms_id)):
            demand_each_firm.append(int(consumption_each_firm[i] // prices[i]))
        real_consumption = await self.environment.economy_client.calculate_consumption(
            firms_id, agent_id, demand_each_firm
        )
        node_id = await self.memory.stream.add(
            topic="economy",
            description=f"I bought some goods, and spent {real_consumption:.1f} on {intention}",
        )
        evaluation = {
            "success": True,
            "evaluation": f"I bought some goods, and spent {real_consumption:.1f} on {intention}",
            "consumed_time": 20,
            "node_id": node_id,
        }
        return evaluation


class EconomyNoneBlock(Block):
    """
    Fallback block for non-economic/non-specified activities.
    """

    name = "EconomyNoneBlock"
    description = "Fallback block for other activities"

    def __init__(self, toolbox: AgentToolbox, agent_memory: Memory):
        super().__init__(toolbox=toolbox, agent_memory=agent_memory)

    async def forward(self, context: DotDict):
        """Log generic activities in economy stream."""
        node_id = await self.memory.stream.add(
            topic="economy", description=f"I {context['current_step']['intention']}"
        )
        return {
            "success": True,
            "evaluation": f'Finished{context["current_step"]["intention"]}',
            "consumed_time": 0,
            "node_id": node_id,
        }


class EconomyBlockParams(BlockParams):
    worktime_estimation_prompt: str = Field(
        default=WORKTIME_ESTIMATE_PROMPT, description="Used to determine the worktime"
    )
    UBI: float = Field(default=0, description="Universal Basic Income")
    num_labor_hours: int = Field(
        default=168, description="Number of labor hours per month"
    )
    productivity_per_labor: float = Field(
        default=1, description="Productivity per labor hour"
    )
    time_diff: int = Field(
        default=30 * 24 * 60 * 60, description="Time difference between two triggers"
    )


class EconomyBlockContext(BlockContext): ...


class EconomyBlock(Block):
    """Orchestrates economic activities through specialized sub-blocks.

    Attributes:
        dispatcher: Routes tasks to appropriate sub-blocks
        work_block: Work activity handler
        consumption_block: Consumption manager
        none_block: Fallback activities
    """

    ParamsType = EconomyBlockParams
    OutputType = SocietyAgentBlockOutput
    ContextType = EconomyBlockContext
    NeedAgent = True
    name = "EconomyBlock"
    description = "Work, shopping, consume, and other economic activities."
    actions = {
        "work": "Support the work action",
        "consume": "Support the consume action",
        "economy_none": "Support other economic operations",
    }

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        block_params: Optional[EconomyBlockParams] = None,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
            block_params=block_params,
        )
        self.month_plan_block = MonthEconomyPlanBlock(
            toolbox=toolbox,
            agent_memory=agent_memory,
            ubi=self.params.UBI,
            num_labor_hours=self.params.num_labor_hours,
            productivity_per_labor=self.params.productivity_per_labor,
            time_diff=self.params.time_diff,
        )
        self.work_block = WorkBlock(
            toolbox=toolbox,
            agent_memory=agent_memory,
            worktime_estimation_prompt=self.params.worktime_estimation_prompt,
        )
        self.consumption_block = ConsumptionBlock(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.none_block = EconomyNoneBlock(toolbox=toolbox, agent_memory=agent_memory)
        self.trigger_time = 0
        self.token_consumption = 0
        self.dispatcher = BlockDispatcher(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.dispatcher.register_blocks(
            [self.work_block, self.consumption_block, self.none_block]
        )

    async def before_forward(self):
        try:
            await self.month_plan_block.forward()
        except Exception as e:
            get_logger().warning(f"EconomyBlock MonthPlanBlock: {e}")
            pass

    async def forward(self, agent_context: DotDict) -> SocietyAgentBlockOutput:
        """Coordinate economic activity execution.

        Workflow:
            1. Use dispatcher to select appropriate handler
            2. Delegate execution to selected block
        """
        try:
            self.trigger_time += 1
            context = agent_context | self.context
            intention = context["current_step"]["intention"]
            selected_block = await self.dispatcher.dispatch(context)
            if selected_block is None:
                return self.OutputType(
                    success=False,
                    evaluation=f"Failed to {intention}",
                    consumed_time=0,
                    node_id=None,
                )
            result = await selected_block.forward(context)
            return self.OutputType(**result)
        except Exception as e:
            get_logger().error(f"EconomyBlock: Error in forward: {e}")
            return self.OutputType(
                success=False,
                evaluation="Failed to forward",
                consumed_time=0,
                node_id=None,
            )


class MonthEconomyPlanBlock(Block):
    """Manages monthly economic planning and mental health assessment.

    Attributes:
        configurable_fields: Economic policy parameters
        economy_client: Interface to economic system
        llm_error: Counter for LLM failures
    """

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Memory,
        ubi: float = 0,
        num_labor_hours: int = 168,
        productivity_per_labor: float = 1,
        time_diff: int = 30 * 24 * 60 * 60,
    ):
        super().__init__(
            toolbox=toolbox,
            agent_memory=agent_memory,
        )
        self.llm_error = 0
        self.last_time_trigger = None
        self.forward_times = 0
        self.ubi = ubi
        self.num_labor_hours = num_labor_hours
        self.productivity_per_labor = productivity_per_labor
        self.time_diff = time_diff

    async def month_trigger(self):
        """Check if monthly planning cycle should activate."""
        now_tick = self.environment.get_tick()
        if (
            self.last_time_trigger is None
            or now_tick - self.last_time_trigger >= self.time_diff
        ):
            self.last_time_trigger = now_tick
            return True
        return False

    async def forward(self):
        """Execute monthly planning workflow.

        Workflow:
            1. Collect economic indicators
            2. Generate LLM prompts for work/consumption propensity
            3. Update agent's economic status
            4. Periodically conduct mental health assessments
            5. Handle UBI policy evaluations
        """
        if await self.month_trigger():
            agent_id = await self.memory.status.get("id")
            firms_id = await self.environment.economy_client.get_firm_ids()
            firm_id = await self.memory.status.get("firm_id")
            bank_id = await self.environment.economy_client.get_bank_ids()
            bank_id = bank_id[0]
            name = await self.memory.status.get("name")
            age = await self.memory.status.get("age")
            city = await self.memory.status.get("city")
            job = await self.memory.status.get("occupation")
            skill, consumption, wealth = await self.environment.economy_client.get(
                agent_id, ["skill", "consumption", "currency"]
            )
            get_logger().debug(f"type of skill: {type(skill)}, value: {skill}")
            tax_paid = await self.memory.status.get("tax_paid")
            prices = await self.environment.economy_client.get(firms_id, "price")
            price = np.mean(prices)
            interest_rate = await self.environment.economy_client.get(
                bank_id, "interest_rate"
            )

            problem_prompt = f"""
                    You're {name}, a {age}-year-old individual living in {city}. As with all Americans, a portion of your monthly income is taxed by the federal government. This taxation system is tiered, income is taxed cumulatively within defined brackets, combined with a redistributive policy: after collection, the government evenly redistributes the tax revenue back to all citizens, irrespective of their earnings.
                """
            job_prompt = f"""
                        In the previous month, you worked as a(an) {job}. If you continue working this month, your expected hourly income will be ${skill:.2f}.
                    """
            consumption_propensity = await self.memory.status.get(
                "consumption_propensity"
            )
            if (consumption <= 0) and (consumption_propensity > 0):
                consumption_prompt = """
                            Besides, you had no consumption due to shortage of goods.
                        """
            else:
                consumption_prompt = f"""
                            Besides, your consumption was ${consumption:.2f}.
                        """
            tax_prompt = f"""Your tax deduction amounted to ${tax_paid:.2f}, and the government uses the tax revenue to provide social services to all citizens."""
            if self.ubi and self.forward_times >= 96:
                tax_prompt = f"{tax_prompt} Specifically, the government directly provides ${self.ubi} per capita in each month."
            price_prompt = f"""Meanwhile, in the consumption market, the average price of essential goods is now at ${price:.2f}."""
            job_prompt = prettify_document(job_prompt)
            obs_prompt = f"""
                            {problem_prompt} {job_prompt} {consumption_prompt} {tax_prompt} {price_prompt}
                            Your current savings account balance is ${wealth:.2f}. Interest rates, as set by your bank, stand at {interest_rate*100:.2f}%. 
                            Your goal is to maximize your utility by deciding how much to work and how much to consume. Your utility is determined by your consumption, income, saving, social service recieved and leisure time. You will spend the time you do not work on leisure activities. 
                            With all these factors in play, and considering aspects like your living costs, any future aspirations, and the broader economic trends, how is your willingness to work this month? Furthermore, how would you plan your expenditures on essential goods, keeping in mind good price?
                            Please share your decisions in a JSON format as follows:
                            {{'work': a value between 0 and 1, indicating the propensity to work,
                            'consumption': a value between 0 and 1, indicating the proportion of all your savings and income you intend to spend on essential goods
                            }}
                            Any other output words are NOT allowed.
                        """
            obs_prompt = prettify_document(obs_prompt)
            try:
                await self.memory.status.update(
                    "dialog_queue",
                    [{"role": "user", "content": obs_prompt}],
                    mode="merge",
                )
                dialog_queue = await self.memory.status.get("dialog_queue")
                content = await self.llm.atext_request(list(dialog_queue), timeout=300)
                await self.memory.status.update(
                    "dialog_queue",
                    [{"role": "assistant", "content": content}],
                    mode="merge",
                )
                propensity_dict = extract_dict_from_string(content)[0]
                work_propensity, consumption_propensity = (
                    propensity_dict["work"],
                    propensity_dict["consumption"],
                )
                if isinstance(work_propensity, numbers.Number) and isinstance(
                    consumption_propensity, numbers.Number
                ):
                    await self.memory.status.update("work_propensity", work_propensity)
                    await self.memory.status.update(
                        "consumption_propensity", consumption_propensity
                    )
                else:
                    self.llm_error += 1
            except Exception:
                self.llm_error += 1

            work_skill = await self.environment.economy_client.get(agent_id, "skill")
            work_propensity = await self.memory.status.get("work_propensity")
            consumption_propensity = await self.memory.status.get(
                "consumption_propensity"
            )
            work_hours = work_propensity * self.num_labor_hours
            # income = await self.economy_client.get(agent_id, 'income')
            income = work_hours * work_skill

            wealth = await self.environment.economy_client.get(agent_id, "currency")
            wealth += work_hours * work_skill
            await self.environment.economy_client.update(agent_id, "currency", wealth)
            await self.environment.economy_client.delta_update_firms(
                firm_id, delta_inventory=int(work_hours * self.productivity_per_labor)
            )

            if self.ubi and self.forward_times >= 96:
                income += self.ubi
                wealth += self.ubi

            await self.memory.status.update(
                "to_consumption_currency", consumption_propensity * wealth
            )

            await self.environment.economy_client.update(agent_id, "consumption", 0)
            await self.environment.economy_client.update(agent_id, "income", income)
            await self.environment.economy_client.update(agent_id, "currency", wealth)

            if self.forward_times % 3 == 0:
                obs_prompt = f"""
                                {problem_prompt} {job_prompt} {consumption_prompt} {tax_prompt} {price_prompt}
                                Your current savings account balance is ${wealth:.2f}. Interest rates, as set by your bank, stand at {interest_rate*100:.2f}%. 
                                Please fill in the following questionnaire:
                                Indicate how often you have felt this way during the last week by choosing one of the following options:
                                "Rarely" means Rarely or none of the time (less than 1 day),
                                "Some" means Some or a little of the time (1-2 days),
                                "Occasionally" means Occasionally or a moderate amount of the time (3-4 days),
                                "Most" means Most or all of the time (5-7 days).
                                Statement 1: I was bothered by things that usually don't bother me.  
                                Statement 2: I did not feel like eating; my appetite was poor.
                                Statement 3: I felt that I could not shake off the blues even with help from my family or friends.
                                Statement 4: I felt that I was just as good as other people.
                                Statement 5: I had trouble keeping my mind on what I was doing.
                                Statement 6: I felt depressed.
                                Statement 7: I felt that everything I did was an effort.
                                Statement 8: I felt hopeful about the future.
                                Statement 9: I thought my life had been a failure.
                                Statement 10: I felt fearful.
                                Statement 11: My sleep was restless.
                                Statement 12: I was happy.
                                Statement 13: I talked less than usual.
                                Statement 14: I felt lonely.
                                Statement 15: People were unfriendly.
                                Statement 16: I enjoyed life.
                                Statement 17: I had crying spells.
                                Statement 18: I felt sad.
                                Statement 19: I felt that people disliked me.
                                Statement 20: I could not get "going".
                                Please response with json format with keys being numbers 1-20 and values being one of "Rarely", "Some", "Occasionally", "Most".
                                Any other output words are NOT allowed.
                            """
                obs_prompt = prettify_document(obs_prompt)
                content = await self.llm.atext_request(
                    [{"role": "user", "content": obs_prompt}], timeout=300
                )
                inverse_score_items = [3, 8, 12, 16]
                category2score = {"rarely": 0, "some": 1, "occasionally": 2, "most": 3}
                try:
                    content = extract_dict_from_string(content)[0]
                    for k in content:
                        if k in inverse_score_items:
                            content[k] = 3 - category2score[content[k].lower()]
                        else:
                            content[k] = category2score[content[k].lower()]
                    depression = sum(list(content.values()))
                    await self.memory.status.update("depression", depression)
                except Exception:
                    self.llm_error += 1

            if self.ubi and self.forward_times >= 96 and self.forward_times % 12 == 0:
                obs_prompt = f"""
                                {problem_prompt} {job_prompt} {consumption_prompt} {tax_prompt} {price_prompt}
                                Your current savings account balance is ${wealth:.2f}. Interest rates, as set by your bank, stand at {interest_rate*100:.2f}%. 
                                What's your opinion on the UBI policy, including the advantages and disadvantages?
                            """
                obs_prompt = prettify_document(obs_prompt)
                content = await self.llm.atext_request(
                    [{"role": "user", "content": obs_prompt}], timeout=300
                )
                await self.memory.status.update("ubi_opinion", [content], mode="merge")

            self.forward_times += 1



from typing import Any, Optional

from pydantic import BaseModel

from ..environment import Environment
from ..llm import LLM
from ..memory import KVMemory, Memory
from .context import BlockContext, DotDict, auto_deepcopy_dotdict, context_to_dot_dict
from .memory_config_generator import MemoryConfig
from .toolbox import AgentToolbox

TRIGGER_INTERVAL = 1

__all__ = [
    "Block",
    "BlockParams",
    "BlockOutput",
]


class BlockParams(BaseModel):
    # TODO: unused
    block_memory: Optional[dict[str, Any]] = None


class BlockOutput(BaseModel): ...


# Behavior Block, used for augmenting agent's behavior capabilities
class Block:
    """
    A foundational component similar to a layer in PyTorch, used for building complex systems.
    """

    ParamsType = BlockParams
    Context = BlockContext
    OutputType = None
    NeedAgent: bool = (
        False  # Whether the block needs an agent, if True, the associated agent will be set automatically
    )
    name: str = ""
    description: str = ""
    actions: dict[str, str] = {}

    def __init__(
        self,
        toolbox: AgentToolbox,
        agent_memory: Optional[Memory] = None,
        block_params: Optional[Any] = None,
    ):
        """
        - **Description**:
            - Initializes a new instance of the Block class with optional LLM, Memory, Simulator, and Trigger components.

        - **Args**:
            - `name` (str): The name of the block.
            - `llm` (Optional[LLM], optional): An instance of LLM. Defaults to None.
            - `environment` (Optional[Environment], optional): An instance of Environment. Defaults to None.
            - `memory` (Optional[Memory], optional): An instance of Memory. Defaults to None.
        """
        self._toolbox = toolbox
        self._agent_memory = agent_memory
        self._agent = None

        # parse block_params
        if block_params is None:
            block_params = self.default_params()
        self.params = block_params
        for key, value in self.params.model_dump().items():
            if key == "block_memory":
                # TODO: remove this after we have a better way to handle block memory
                if isinstance(value, MemoryConfig):
                    self._block_memory = KVMemory(
                        value,
                        embedding=self._toolbox.embedding,
                    )
            else:
                setattr(self, key, value)

        # initialize context
        context = self.default_context()
        self.context = context_to_dot_dict(context)

    @classmethod
    def default_params(cls) -> ParamsType:
        return cls.ParamsType()

    @classmethod
    def default_context(cls) -> Context:
        return cls.Context()

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Create a new dictionary that inherits from parent
        cls.get_functions = dict(cls.__base__.get_functions) if hasattr(cls.__base__, "get_functions") else {}  # type: ignore

        # Register all methods with _register_info
        for name, method in cls.__dict__.items():
            if hasattr(method, "_register_info"):
                info = method._register_info
                cls.get_functions[info["function_name"]] = info

    def set_agent(self, agent: Any):
        self._agent = agent

    @property
    def agent(self) -> Any:
        if self._agent is None:
            raise RuntimeError(
                "Agent access before assignment, please `set_agent` first!"
            )
        return self._agent

    @property
    def toolbox(self) -> AgentToolbox:
        return self._toolbox

    @property
    def llm(self) -> LLM:
        return self._toolbox.llm

    @property
    def memory(self) -> Memory:
        if self._agent_memory is None:
            raise RuntimeError(
                "Memory access before assignment, please `set_memory` first!"
            )
        return self._agent_memory

    @property
    def agent_memory(self) -> Memory:
        if self._agent_memory is None:
            raise RuntimeError(
                "Memory access before assignment, please `set_memory` first!"
            )
        return self._agent_memory

    @property
    def block_memory(self) -> KVMemory:
        if self._block_memory is None:
            raise RuntimeError(
                "Block memory access before assignment, please `set_block_memory` first!"
            )
        return self._block_memory

    @property
    def environment(self) -> Environment:
        return self._toolbox.environment

    async def before_forward(self):
        """
        Before forward - prepare context
        """
        pass

    async def after_forward(self):
        """
        After forward - update/clean context
        """
        pass

    @auto_deepcopy_dotdict
    async def forward(self, agent_context: DotDict):
        """
        - **Description**:
            - Each block performs a specific reasoning task. This method should be overridden by subclasses.
            - Subclasses can define their own parameters as needed.

        - **Args**:
            - `*args`: Variable length argument list.
            - `**kwargs`: Arbitrary keyword arguments.

        - **Raises**:
            - `NotImplementedError`: Subclasses must implement this method.
        """
        raise NotImplementedError("Subclasses should implement this method")



from pydantic import BaseModel

from ...storage import metric

__all__ = ["metric", "ApiMetric"]

# API Models

class ApiMetric(BaseModel):
    """Metric model for API"""

    key: str
    """Metric key"""
    value: float
    """Metric value"""
    step: int
    """Step number"""

    class Config:
        from_attributes = True



import base64

__all__ = ["encode_to_base64"]


def encode_to_base64(input_string: str) -> str:
    # 将字符串转换为字节
    input_bytes = input_string.encode("utf-8")

    # 将字节编码为Base64
    base64_bytes = base64.b64encode(input_bytes)

    # 将Base64字节转换为字符串
    base64_string = base64_bytes.decode("utf-8")

    return base64_string



import asyncio
import os
import random
import time
from enum import Enum
from multiprocessing import cpu_count
from typing import Any, List, Optional, Union, overload

import httpx
import ray
from openai import NOT_GIVEN, APIConnectionError, AsyncOpenAI, NotGiven, OpenAIError
from openai.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    completion_create_params,
)
from pydantic import BaseModel, Field, field_serializer, model_validator

from ..logger import get_logger

os.environ["GRPC_VERBOSITY"] = "ERROR"

__all__ = [
    "LLM",
    "LLMConfig",
    "LLMProviderType",
]

MAX_TIMEOUT = 60

class LLMProviderType(str, Enum):
    """
    Defines the types of LLM providers.
    - **Description**:
        - Enumerates different types of LLM providers.

    - **Types**:
        - `OPENAI`: OpenAI and compatible providers (based on base_url).
        - `DEEPSEEK`: DeepSeek.
        - `QWEN`: Qwen.
        - `ZHIPU`: Zhipu.
        - `SILICONFLOW`: SiliconFlow.
        - `VLLM`: VLLM.
    """

    OpenAI = "openai"
    DeepSeek = "deepseek"
    Qwen = "qwen"
    ZhipuAI = "zhipuai"
    SiliconFlow = "siliconflow"
    VolcEngine = "volcengine"
    VLLM = "vllm"


class LLMConfig(BaseModel):
    """LLM configuration class."""

    provider: LLMProviderType = Field(...)
    """The type of the LLM provider"""

    base_url: Optional[str] = Field(None)
    """The base URL for the LLM provider"""

    api_key: str = Field(...)
    """API key for accessing the LLM provider"""

    model: str = Field(...)
    """The model to use"""

    concurrency: int = Field(200, ge=1)
    """Concurrency value for LLM operations to avoid rate limit"""

    timeout: float = Field(30, ge=1, le=MAX_TIMEOUT)
    """Timeout for LLM operations in seconds"""

    @field_serializer("provider")
    def serialize_provider(self, provider: LLMProviderType, info):
        return provider.value

    @model_validator(mode="after")
    def validate_configuration(self):
        if self.provider != LLMProviderType.VLLM and self.base_url is not None:
            raise ValueError("base_url is not supported for this provider")
        return self


@ray.remote
class LLMActor:
    """
    Actor class for LLM operations.
    """

    def __init__(self):
        self._http_client = httpx.AsyncClient(
            timeout=httpx.Timeout(
                connect=min(30.0, MAX_TIMEOUT / 4),  # 连接超时时间
                read=MAX_TIMEOUT,  # 读取超时时间
                write=MAX_TIMEOUT,  # 写入超时时间
                pool=MAX_TIMEOUT,  # 连接池超时时间
            ),
            limits=httpx.Limits(
                max_keepalive_connections=20, max_connections=100, keepalive_expiry=30.0
            ),
        )

    async def call(
        self,
        config: LLMConfig,
        dialog: list[ChatCompletionMessageParam],
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        temperature: float = 1,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        timeout: int = 300,
        retries: int = 10,
        tools: Union[List[ChatCompletionToolParam], NotGiven] = NOT_GIVEN,
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven] = NOT_GIVEN,
    ):
        """
        Sends an asynchronous text request to the configured LLM API.

        - **Description**:
            - Attempts to send a text request up to `retries` times with exponential backoff on failure.
            - Handles different request types and manages token usage statistics.

        - **Args**:
            - `config`: LLM configuration.
            - `dialog`: Messages to send as part of the chat completion request.
            - `response_format`: JSON schema for the response. Default is NOT_GIVEN.
            - `temperature`: Controls randomness in the model's output. Default is 1.
            - `max_tokens`: Maximum number of tokens to generate in the response. Default is None.
            - `top_p`: Limits the next token selection to a subset of tokens with a cumulative probability above this value. Default is None.
            - `frequency_penalty`: Penalizes new tokens based on their existing frequency in the text so far. Default is None.
            - `presence_penalty`: Penalizes new tokens based on whether they appear in the text so far. Default is None.
            - `timeout`: Request timeout in seconds. Default is 300 seconds.
            - `retries`: Number of retry attempts in case of failure. Default is 10.
            - `tools`: List of dictionaries describing the tools that can be called by the model. Default is NOT_GIVEN.
            - `tool_choice`: Dictionary specifying how the model should choose from the provided tools. Default is NOT_GIVEN.

        - **Returns**:
            - A string containing the message content or a dictionary with tool call arguments if tools are used.
            - Raises exceptions if the request fails after all retry attempts.
        """
        start_time = time.time()

        log = {
            "request_time": start_time,
            "total_errors": 0,
            "error_types": {
                "connection_error": 0,
                "openai_error": 0,
                "timeout_error": 0,
                "other_error": 0,
            },
            "input_tokens": 0,
            "output_tokens": 0,
        }

        client = AsyncOpenAI(
            api_key=config.api_key,
            timeout=config.timeout,
            base_url=config.base_url,
            http_client=self._http_client,
        )
        for attempt in range(retries):
            response = None
            try:
                response = await client.chat.completions.create(
                    model=config.model,
                    messages=dialog,
                    response_format=response_format,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    frequency_penalty=frequency_penalty,
                    presence_penalty=presence_penalty,
                    stream=False,
                    timeout=timeout,
                    tools=tools,
                    tool_choice=tool_choice,
                )
                if response.usage is not None:
                    log["input_tokens"] += response.usage.prompt_tokens
                    log["output_tokens"] += response.usage.completion_tokens
                else:
                    get_logger().warning(f"No usage in response: {response}")
                end_time = time.time()
                log["consumption"] = end_time - start_time
                if tools:
                    return response, log
                else:
                    content = response.choices[0].message.content
                    if content is None:
                        raise ValueError("No content in response")
                    return content, log
            except APIConnectionError as e:
                get_logger().warning(
                    f"API connection error: `{e}` for request {dialog} {tools} {tool_choice}. original response: `{response}`. Retry {attempt+1} of {retries}"
                )
                log["total_errors"] += 1
                log["error_types"]["connection_error"] += 1
                if attempt < retries - 1:
                    time.sleep(random.random() * 2**attempt)
                else:
                    raise e
            except OpenAIError as e:
                error_message = str(e)
                get_logger().warning(
                    f"OpenAIError: {error_message} for request {dialog} {tools} {tool_choice}. original response: `{response}`. Retry {attempt+1} of {retries}"
                )
                log["total_errors"] += 1
                log["error_types"]["openai_error"] += 1
                if attempt < retries - 1:
                    time.sleep(random.random() * 2**attempt)
                else:
                    raise e
            except asyncio.TimeoutError as e:
                get_logger().warning(
                    f"TimeoutError: `{e}` for request {dialog} {tools} {tool_choice}. original response: `{response}`. Retry {attempt+1} of {retries}"
                )
                log["total_errors"] += 1
                log["error_types"]["timeout_error"] += 1
                if attempt < retries - 1:
                    time.sleep(random.random() * 2**attempt)
                else:
                    raise e
            except Exception as e:
                get_logger().warning(
                    f"LLM Error: `{e}` for request {dialog} {tools} {tool_choice}. original response: `{response}`. Retry {attempt+1} of {retries}"
                )
                log["total_errors"] += 1
                log["error_types"]["other_error"] += 1
                if attempt < retries - 1:
                    time.sleep(random.random() * 2**attempt)
                else:
                    raise e
        raise RuntimeError("Failed to get response from LLM")


class LLM:
    """
    Main class for the Large Language Model (LLM) object used by Agent(Soul).

    - **Description**:
        - This class manages configurations and interactions with different large language model APIs.
        - It initializes clients based on the specified request type and handles token usage and consumption reporting.
    """

    def __init__(self, configs: List[LLMConfig], num_actors: int = min(cpu_count(), 8)):
        """
        Initializes the LLM instance.

        - **Parameters**:
            - `configs`: An instance of `LLMConfig` containing configuration settings for the LLM.
            - `num_actors` (`int`): Number of actor instances to create for parallel processing. Defaults to min(cpu_count(), 8).
        """

        if len(configs) == 0:
            raise ValueError(
                "No LLM config is provided, please check your configuration"
            )

        self.configs = configs
        self._semaphores = [
            asyncio.Semaphore(config.concurrency) for config in self.configs
        ]
        self._log_list = []
        self.prompt_tokens_used = 0
        self.completion_tokens_used = 0
        self._next_index = 0
        self._last_show_time = time.time()

        for config in self.configs:
            base_url = config.base_url
            if base_url is not None:
                base_url = base_url.rstrip("/")

            if config.provider == LLMProviderType.OpenAI:
                ...
            elif config.provider == LLMProviderType.DeepSeek:
                base_url = "https://api.deepseek.com/v1"
            elif config.provider == LLMProviderType.Qwen:
                base_url = "https://dashscope.aliyuncs.com/compatible-mode/v1"
            elif config.provider == LLMProviderType.SiliconFlow:
                base_url = "https://api.siliconflow.cn/v1"
            elif config.provider == LLMProviderType.VLLM:
                ...
            elif config.provider == LLMProviderType.ZhipuAI:
                base_url = "https://open.bigmodel.cn/api/paas/v4/"
            elif config.provider == LLMProviderType.VolcEngine:
                base_url = "https://ark.cn-beijing.volces.com/api/v3/"
            else:
                raise ValueError(f"Unsupported `provider` {config.provider}!")
            config.base_url = base_url

        self._actors = [LLMActor.remote() for _ in range(num_actors)]
        self._lock = asyncio.Lock()

    def get_log_list(self):
        return self._log_list

    def clear_log_list(self):
        self._log_list = []

    def _get_index(self):
        self._next_index += 1
        return self._next_index

    @overload
    async def atext_request(
        self,
        dialog: list[ChatCompletionMessageParam],
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        temperature: float = 1,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        timeout: int = 300,
        retries: int = 10,
        tools: NotGiven = NOT_GIVEN,
        tool_choice: NotGiven = NOT_GIVEN,
    ) -> str: ...

    @overload
    async def atext_request(
        self,
        dialog: list[ChatCompletionMessageParam],
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        temperature: float = 1,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        timeout: int = 300,
        retries: int = 10,
        tools: List[ChatCompletionToolParam] = [],
        tool_choice: ChatCompletionToolChoiceOptionParam = "auto",
    ) -> Any: ...

    async def atext_request(
        self,
        dialog: list[ChatCompletionMessageParam],
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        temperature: float = 1,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        timeout: int = 300,
        retries: int = 10,
        tools: Union[List[ChatCompletionToolParam], NotGiven] = NOT_GIVEN,
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven] = NOT_GIVEN,
    ):
        """
        Sends an asynchronous text request to the configured LLM API.

        - **Description**:
            - Attempts to send a text request up to `retries` times with exponential backoff on failure.
            - Handles different request types and manages token usage statistics.

        - **Args**:
            - `dialog`: Messages to send as part of the chat completion request.
            - `response_format`: JSON schema for the response. Default is NOT_GIVEN.
            - `temperature`: Controls randomness in the model's output. Default is 1.
            - `max_tokens`: Maximum number of tokens to generate in the response. Default is None.
            - `top_p`: Limits the next token selection to a subset of tokens with a cumulative probability above this value. Default is None.
            - `frequency_penalty`: Penalizes new tokens based on their existing frequency in the text so far. Default is None.
            - `presence_penalty`: Penalizes new tokens based on whether they appear in the text so far. Default is None.
            - `timeout`: Request timeout in seconds. Default is 300 seconds.
            - `retries`: Number of retry attempts in case of failure. Default is 10.
            - `tools`: List of dictionaries describing the tools that can be called by the model. Default is NOT_GIVEN.
            - `tool_choice`: Dictionary specifying how the model should choose from the provided tools. Default is NOT_GIVEN.

        - **Returns**:
            - A string containing the message content or a dictionary with tool call arguments if tools are used.
            - Raises exceptions if the request fails after all retry attempts.
        """
        index = self._get_index()
        client_i = index % len(self.configs)
        actor_i = index % len(self._actors)
        if time.time() - self._last_show_time > 10:
            get_logger().info(
                f"LLM request count: {index}"
            )
            self._last_show_time = time.time()
        async with self._semaphores[client_i]:
            content, log = await self._actors[actor_i].call.remote(  # type: ignore
                self.configs[client_i],
                dialog,
                response_format,
                temperature,
                max_tokens,
                top_p,
                frequency_penalty,
                presence_penalty,
                timeout,
                retries,
                tools,
                tool_choice,
            )
            self._log_list.append(log)
            self.prompt_tokens_used += log["input_tokens"]
            self.completion_tokens_used += log["output_tokens"]
        return content



import enum
from typing import Optional
from pydantic import BaseModel
from datetime import datetime
import uuid

__all__ = [
    "StorageSurvey",
    "StorageDialogType",
    "StorageDialog",
    "StorageGlobalPrompt",
    "StorageProfile",
    "StorageStatus",
    "StoragePendingDialog",
    "StoragePendingSurvey",
    "StorageTaskResult",
]


class StorageExpInfo(BaseModel):
    tenant_id: str
    id: str
    name: str
    num_day: int
    status: int
    cur_day: int
    cur_t: float
    config: str
    error: str
    input_tokens: int
    output_tokens: int
    created_at: datetime
    updated_at: datetime


class StorageSurvey(BaseModel):
    id: int
    day: int
    t: float
    survey_id: uuid.UUID
    result: str
    created_at: datetime


class StorageDialogType(enum.IntEnum):
    """Storage dialog type"""

    Thought = 0  # Dialog in agent self
    Talk = 1  # Dialog with other agents
    User = 2  # Dialog with user


class StorageDialog(BaseModel):
    id: int
    day: int
    t: float
    type: int
    speaker: str
    content: str
    created_at: datetime


class StorageGlobalPrompt(BaseModel):
    day: int
    t: float
    prompt: str
    created_at: datetime


class StorageProfile(BaseModel):
    id: int
    name: str
    profile: str


class StorageStatus(BaseModel):
    id: int
    day: int
    t: float
    lng: Optional[float]
    lat: Optional[float]
    parent_id: Optional[int]
    action: str
    status: str
    created_at: datetime


class StoragePendingDialog(BaseModel):
    """Pending dialog storage type"""

    id: int
    """Pending dialog ID"""
    agent_id: int
    """Agent ID"""
    day: int
    """Day"""
    t: float
    """Time"""
    content: str
    """Content"""
    created_at: datetime
    """Created time"""
    processed: bool
    """Whether the dialog has been processed"""


class StoragePendingSurvey(BaseModel):
    """Pending survey storage type"""

    id: int
    """Pending survey ID"""
    agent_id: int
    """Agent ID"""
    day: int
    """Day"""
    t: float
    """Time"""
    survey_id: uuid.UUID
    """Survey ID"""
    data: dict
    """Data"""
    created_at: datetime
    """Created time"""
    processed: bool
    """Whether the survey has been processed"""


class StorageTaskResult(BaseModel):
    id: int
    """Task ID"""
    agent_id: int
    """Agent ID"""
    context: dict
    """Context"""
    ground_truth: dict
    """Ground truth"""
    result: dict
    """Result"""
    created_at: datetime



"""Simulator: Urban Simulator"""

import asyncio
import os
import tempfile
from datetime import datetime
from subprocess import Popen
from typing import Any, Literal, Optional, Tuple, Union, overload

import yaml
from pycityproto.city.map.v2 import map_pb2 as map_pb2
from pycityproto.city.person.v2 import person_pb2 as person_pb2
from pycityproto.city.person.v2 import person_service_pb2 as person_service
from pycityproto.city.trip.v2.trip_pb2 import TripMode
from pydantic import BaseModel, ConfigDict, Field
from pyproj import Proj
from shapely.geometry import Point

from ..logger import get_logger
from ..s3 import S3Client, S3Config
from .economy.econ_client import EconomyClient
from .mapdata import MapConfig, MapData
from .sim import CityClient
from .syncer import Syncer
from .utils import find_free_ports
from .utils.base64 import encode_to_base64
from .utils.const import POI_CATG_DICT
from .utils.protobuf import dict2pb
from .download_sim import download_binary

__all__ = [
    "Environment",
    "EnvironmentStarter",
    "EnvironmentConfig",
]


POI_START_ID = 7_0000_0000


class EnvironmentConfig(BaseModel):
    """Configuration for the simulation environment."""

    model_config = ConfigDict(use_enum_values=True, use_attribute_docstrings=True)

    start_tick: int = Field(default=8 * 60 * 60)
    """Starting tick of one day, in seconds"""

    metric_interval: int = Field(default=3600)
    """Interval of metrics, in ticks"""

    weather: str = Field(default="The weather is sunny")
    """Current weather condition in the environment"""

    temperature: str = Field(default="The temperature is 23C")
    """Current temperature in the environment"""

    workday: bool = Field(default=True)
    """Indicates if it's a working day"""

    other_information: str = Field(default="")
    """Additional environment information"""

    def to_prompts(self) -> dict[str, Any]:
        """Convert the environment config to prompts"""
        return {
            "weather": self.weather,
            "temperature": self.temperature,
            "workday": self.workday,
            "other_information": self.other_information,
        }


class Environment:
    """
    The environment, including map data, simulator clients, and environment variables.
    """

    def __init__(
        self,
        map_data: MapData,
        server_addr: Optional[str],
        environment_config: EnvironmentConfig,
        citizen_ids: set[int] = set(),
        firm_ids: set[int] = set(),
        bank_ids: set[int] = set(),
        nbs_ids: set[int] = set(),
        government_ids: set[int] = set(),
    ):
        """
        Initialize the Environment.

        - **Args**:
            - `map_data`: `MapData`, map data
            - `server_addr`: `str`, server address
            - `environment_config`: `EnvironmentConfig`, environment config
            - `citizen_ids`: `set[int]`, citizen ids
            - `firm_ids`: `set[int]`, firm ids
            - `bank_ids`: `set[int]`, bank ids
            - `nbs_ids`: `set[int]`, nbs ids
            - `government_ids`: `set[int]`, government ids
            - `syncer`: `ray.ObjectRef`, syncer for get_tick
        """
        self._map = map_data
        self._create_poi_id_2_aoi_id()
        self._server_addr = server_addr
        self.poi_cate = POI_CATG_DICT
        """poi categories"""

        self._projector = Proj(self._map.get_projector())
        self._environment_config = environment_config
        self._environment_prompt = environment_config.to_prompts()

        self._log_list = []
        """log list"""

        self._lock = asyncio.Lock()
        """lock for simulator"""

        self._tick = 0
        """number of simulated ticks"""

        self._aoi_message: dict[int, dict[int, list[str]]] = {}
        """aoi message"""

        self._init_citizen_ids = citizen_ids
        self._init_firm_ids = firm_ids
        self._init_bank_ids = bank_ids
        self._init_nbs_ids = nbs_ids
        self._init_government_ids = government_ids

        # type annotation
        self._server_addr = server_addr
        self._city_client: Optional[CityClient] = None
        self._economy_client: Optional[EconomyClient] = None

    def init(self) -> Any:
        assert self._server_addr is not None, "Server address not initialized"
        self._city_client = CityClient(self._server_addr)
        self._economy_client = EconomyClient(self._server_addr)
        self._economy_client.set_ids(
            citizen_ids=self._init_citizen_ids,
            firm_ids=self._init_firm_ids,
            bank_ids=self._init_bank_ids,
            nbs_ids=self._init_nbs_ids,
            government_ids=self._init_government_ids,
        )

    def close(self) -> Any:
        """Close the Environment."""
        pass

    def _create_poi_id_2_aoi_id(self):
        assert self._map is not None
        pois = self._map.get_all_pois()
        self.poi_id_2_aoi_id: dict[int, int] = {
            poi["id"]: poi["aoi_id"] for poi in pois
        }

    @property
    def map(self):
        assert self._map is not None, "Map not initialized"
        return self._map

    @property
    def city_client(self):
        assert self._city_client is not None, "Client not initialized"
        return self._city_client

    @property
    def economy_client(self):
        assert self._economy_client is not None, "Economy client not initialized"
        return self._economy_client

    @property
    def projector(self):
        return self._projector

    def get_log_list(self):
        return self._log_list

    def clear_log_list(self):
        self._log_list = []

    def get_poi_cate(self):
        return self.poi_cate

    def get_aoi_ids(self):
        aois = self._map.get_all_aois()
        return [aoi["id"] for aoi in aois]

    def register_aoi_message(
        self, agent_id: int, target_aoi: Union[int, list[int]], content: str
    ):
        """
        Register aoi message

        - **Args**:
            - `target_aoi` (`Union[int, list[int]]`): The ID of the target aoi.
            - `content` (`str`): The content of the message to send.
        """
        if isinstance(target_aoi, int):
            target_aoi = [target_aoi]
        for aoi in target_aoi:
            if aoi not in self._aoi_message:
                self._aoi_message[aoi] = {}
            if agent_id not in self._aoi_message[aoi]:
                self._aoi_message[aoi][agent_id] = []
            self._aoi_message[aoi][agent_id].append(content)

    def cancel_aoi_message(self, agent_id: int, target_aoi: Union[int, list[int]]):
        """
        Cancel aoi message
        """
        if isinstance(target_aoi, int):
            target_aoi = [target_aoi]
        for aoi in target_aoi:
            if aoi in self._aoi_message:
                if agent_id in self._aoi_message[aoi]:
                    self._aoi_message[aoi].pop(agent_id)
                if len(self._aoi_message[aoi]) == 0:
                    self._aoi_message.pop(aoi)

    @property
    def environment(self) -> dict[str, str]:
        """
        Get the current state of environment variables.
        """
        return self._environment_prompt

    def set_environment(self, environment: dict[str, str]):
        """
        Set the entire dictionary of environment variables.

        - **Args**:
            - `environment` (`Dict[str, str]`): Key-value pairs of environment variables.
        """
        self._environment_prompt = environment

    def sense(self, key: str) -> Any:
        """
        Retrieve the value of an environment variable by its key.

        - **Args**:
            - `key` (`str`): The key of the environment variable.

        - **Returns**:
            - `Any`: The value of the corresponding key, or an empty string if not found.
        """
        return self._environment_prompt.get(key, "Don't know")

    def sense_aoi(self, aoi_id: int) -> str:
        """
        Retrieve the value of an environment variable by its key.
        """
        if aoi_id in self._aoi_message:
            msg_ = ""
            aoi_messages = self._aoi_message[aoi_id]
            for _, messages in aoi_messages.items():
                for message in messages:
                    msg_ += message + "\n"
            return msg_
        else:
            return ""

    def update_environment(self, key: str, value: Any):
        """
        Update the value of a single environment variable.

        - **Args**:
            - `key` (`str`): The key of the environment variable.
            - `value` (`Any`): The new value to set.
        """
        self._environment_prompt[key] = value

    def get_environment(self) -> str:
        global_prompt = ""
        for key in self._environment_prompt:
            value = self._environment_prompt[key]
            if isinstance(value, str):
                global_prompt += f"{key}: {value}\n"
            elif isinstance(value, dict):
                for k, v in value.items():
                    global_prompt += f"{key}.{k}: {v}\n"
            elif isinstance(value, bool):
                global_prompt += f"Is it {key}: {value}\n"
            elif isinstance(value, list):
                global_prompt += f"{key} elements: {value}\n"
            else:
                global_prompt += f"{key}: {value}\n"
        return global_prompt

    def get_poi_categories(
        self,
        center: Optional[Union[tuple[float, float], Point]] = None,
        radius: Optional[float] = None,
    ) -> list[str]:
        """
        Retrieve unique categories of Points of Interest (POIs) around a central point.

        - **Args**:
            - `center` (`Optional[Union[Tuple[float, float], Point]]`): The central point as a tuple or Point object.
              Defaults to (0, 0) if not provided.
            - `radius` (`Optional[float]`): The search radius in meters. If not provided, all POIs are considered.

        - **Returns**:
            - `List[str]`: A list of unique POI category names.
        """
        categories: list[str] = []
        if center is None:
            center = (0, 0)
        assert self._map is not None
        _pois: list[Any] = self._map.query_pois(
            center=center,
            radius=radius,
            return_distance=False,
        )
        for poi in _pois:
            catg = poi["category"]
            categories.append(catg.split("|")[-1])
        return list(set(categories))

    def set_tick(self, tick: int):
        self._tick = tick

    def get_tick(self) -> int:
        return self._tick

    @overload
    def get_datetime(self) -> tuple[int, int]: ...
    @overload
    def get_datetime(self, format_time: Literal[False]) -> tuple[int, int]: ...
    @overload
    def get_datetime(
        self, format_time: Literal[True], format: str = "%H:%M:%S"
    ) -> tuple[int, str]: ...

    def get_datetime(
        self, format_time: bool = False, format: str = "%H:%M:%S"
    ) -> Union[tuple[int, int], tuple[int, str]]:
        """
        Get the current time of the simulator.

        By default, returns the number of seconds since midnight. Supports formatted output.

        - **Args**:
            - `format_time` (`bool`): Whether to return the time in a formatted string. Defaults to `False`.
            - `format` (`str`): The format string for formatting the time. Defaults to "%H:%M:%S".

        - **Returns**:
            - `Union[tuple[int, int], tuple[int, str]]`: The current simulation (day, time) either as an integer representing seconds since midnight or as a formatted string.
        """

        tick = self._tick
        day = (tick + self._environment_config.start_tick) // (24 * 60 * 60)
        time = (tick + self._environment_config.start_tick) % (24 * 60 * 60)
        if format_time:
            hours = time // 3600
            minutes = (time % 3600) // 60
            seconds = time % 60
            formatted_time = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
            if format != "%H:%M:%S":
                temp_time = datetime.strptime(formatted_time, "%H:%M:%S")
                formatted_time = temp_time.strftime(format)
            return (day, formatted_time)
        else:
            return (day, time)

    async def get_person(self, person_id: int) -> dict:
        """
        Retrieve information about a specific person by ID.

        - **Args**:
            - `person_id` (`int`): The ID of the person to retrieve information for.

        - **Returns**:
            - `Dict`: Information about the specified person.
        """
        person = await self.city_client.person_service.GetPerson(
            req={"person_id": person_id}
        )
        return person

    async def add_person(self, dict_person: dict) -> dict:
        """
        Add a new person to the simulation.

        - **Args**:
            - `dict_person` (`dict`): The person object to add.

        - **Returns**:
            - `Dict`: Response from adding the person.
        """
        person = dict2pb(dict_person, person_pb2.Person())
        if isinstance(person, person_pb2.Person):
            req = person_service.AddPersonRequest(person=person)
        else:
            req = person
        resp: dict = await self.city_client.person_service.AddPerson(req)
        return resp

    async def set_aoi_schedules(
        self,
        person_id: int,
        target_positions: Union[
            list[Union[int, tuple[int, int]]], Union[int, tuple[int, int]]
        ],
        departure_times: Optional[list[float]] = None,
        modes: Optional[list[TripMode]] = None,
    ):
        """
        Set schedules for a person to visit Areas of Interest (AOIs).

        - **Args**:
            - `person_id` (`int`): The ID of the person whose schedule is being set.
            - `target_positions` (`Union[List[Union[int, Tuple[int, int]]], Union[int, Tuple[int, int]]]`):
              A list of AOI or POI IDs or tuples of (AOI ID, POI ID) that the person will visit.
            - `departure_times` (`Optional[List[float]]`): Departure times for each trip in the schedule.
              If not provided, current time will be used for all trips.
            - `modes` (`Optional[List[int]]`): Travel modes for each trip.
              Defaults to `TRIP_MODE_DRIVE_ONLY` if not specified.
        """
        cur_time = float(self.get_tick())
        if not isinstance(target_positions, list):
            target_positions = [target_positions]
        if departure_times is None:
            departure_times = [cur_time for _ in range(len(target_positions))]
        else:
            for _ in range(len(target_positions) - len(departure_times)):
                departure_times.append(cur_time)
        if modes is None:
            modes = [
                TripMode.TRIP_MODE_DRIVE_ONLY for _ in range(len(target_positions))
            ]
        else:
            for _ in range(len(target_positions) - len(modes)):
                modes.append(TripMode.TRIP_MODE_DRIVE_ONLY)
        _schedules = []
        for target_pos, _time, _mode in zip(target_positions, departure_times, modes):
            if isinstance(target_pos, int):
                if target_pos >= POI_START_ID:
                    poi_id = target_pos
                    end = {
                        "aoi_position": {
                            "aoi_id": self.poi_id_2_aoi_id[poi_id],
                            "poi_id": poi_id,
                        }
                    }
                else:
                    aoi_id = target_pos
                    end = {
                        "aoi_position": {
                            "aoi_id": aoi_id,
                        }
                    }
            else:
                aoi_id, poi_id = target_pos
                end = {"aoi_position": {"aoi_id": aoi_id, "poi_id": poi_id}}
                # activity = ""
            trips = [
                {
                    "mode": _mode,
                    "end": end,
                    "departure_time": _time,
                },
            ]
            _schedules.append(
                {"trips": trips, "loop_count": 1, "departure_time": _time}
            )
        req = {"person_id": person_id, "schedules": _schedules}
        await self.city_client.person_service.SetSchedule(req)

    async def reset_person_position(
        self,
        person_id: int,
        aoi_id: Optional[int] = None,
        poi_id: Optional[int] = None,
        lane_id: Optional[int] = None,
        s: Optional[float] = None,
    ):
        """
        Reset the position of a person within the simulation.

        - **Args**:
            - `person_id` (`int`): The ID of the person whose position is being reset.
            - `aoi_id` (`Optional[int]`): The ID of the Area of Interest (AOI) where the person should be placed.
            - `poi_id` (`Optional[int]`): The ID of the Point of Interest (POI) within the AOI.
            - `lane_id` (`Optional[int]`): The ID of the lane on which the person should be placed.
            - `s` (`Optional[float]`): The longitudinal position along the lane.
        """
        reset_position = {}
        if aoi_id is not None:
            reset_position["aoi_position"] = {"aoi_id": aoi_id}
            if poi_id is not None:
                reset_position["aoi_position"]["poi_id"] = poi_id
            get_logger().debug(
                f"Setting person {person_id} pos to AoiPosition {reset_position}"
            )
            await self.city_client.person_service.ResetPersonPosition(
                {"person_id": person_id, "position": reset_position}
            )
        elif lane_id is not None:
            reset_position["lane_position"] = {
                "lane_id": lane_id,
                "s": 0.0,
            }
            if s is not None:
                reset_position["lane_position"]["s"] = s
            get_logger().debug(
                f"Setting person {person_id} pos to LanePosition {reset_position}"
            )
            await self.city_client.person_service.ResetPersonPosition(
                {"person_id": person_id, "position": reset_position}
            )
        else:
            get_logger().debug(
                f"Neither aoi or lane pos provided for person {person_id} position reset!!"
            )

    def get_around_poi(
        self,
        center: Union[tuple[float, float], Point],
        radius: float,
        poi_type: Union[str, list[str]],
    ) -> list[dict]:
        """
        Get Points of Interest (POIs) around a central point based on type.

        - **Args**:
            - `center` (`Union[Tuple[float, float], Point]`): The central point as a tuple or Point object.
            - `radius` (`float`): The search radius in meters.
            - `poi_type` (`Union[str, List[str]]`): The category or categories of POIs to filter by.

        - **Returns**:
            - `List[Dict]`: A list of dictionaries containing information about the POIs found.
        """
        if isinstance(poi_type, str):
            poi_type = [poi_type]
        transformed_poi_type: list[str] = []
        for t in poi_type:
            if t not in self.poi_cate:
                transformed_poi_type.append(t)
            else:
                transformed_poi_type += self.poi_cate[t]
        poi_type_set = set(transformed_poi_type)
        # query pois within the radius
        assert self._map is not None
        _pois: list[dict] = self._map.query_pois(
            center=center,
            radius=radius,
            return_distance=False,
        )
        # Filter out POIs that do not meet the category prefix
        pois = []
        for poi in _pois:
            catg = poi["category"]
            if catg.split("|")[-1] not in poi_type_set:
                continue
            pois.append(poi)
        return pois


class EnvironmentStarter(Environment):
    """
    The entrypoint of the simulator, used to initialize the simulator and map.

    - **Description**:
        - This class is the core of the simulator, responsible for initializing and managing the simulation environment.
        - It reads parameters from a configuration dictionary, initializes map data, and starts or connects to a simulation server as needed.
    """

    def __init__(
        self,
        map_config: MapConfig,
        environment_config: EnvironmentConfig,
        s3config: S3Config,
        log_dir: str,
        home_dir: str,
    ):
        """
        Environment config

        - **Args**:
            - `map_config` (MapConfig): Map config
            - `simulator_config` (SimulatorConfig): Simulator config
            - `environment_config` (EnvironmentConfig): Environment config
        """
        self._sim_bin_path = download_binary(home_dir)
        self._map_config = map_config
        self._environment_config = environment_config
        self._s3config = s3config
        self._log_dir = log_dir
        self._home_dir = home_dir
        mapdata = MapData(map_config, s3config)

        super().__init__(mapdata, None, environment_config)

        self._last_metric_tick = -1e999

        # type annotation
        self._sim_proc: Optional[Popen] = None

    def to_init_args(self):
        return {
            "map_data": self._map,
            "server_addr": self._server_addr,
            "environment_config": self._environment_config,
            "citizen_ids": self.economy_client._citizen_ids,
            "firm_ids": self.economy_client._firm_ids,
            "bank_ids": self.economy_client._bank_ids,
            "nbs_ids": self.economy_client._nbs_ids,
            "government_ids": self.economy_client._government_ids,
        }

    @property
    def syncer(self):
        assert self._syncer is not None, "Syncer not initialized"
        return self._syncer

    async def init(self):
        """
        Initialize the environment including the syncer and the simulator.
        """
        # =========================
        # init syncer
        # =========================
        sim_port, syncer_port = find_free_ports(2)
        syncer_addr = f"localhost:{syncer_port}"
        self._syncer = Syncer(syncer_addr)
        await self._syncer.init()

        # =========================
        # init simulator
        # =========================

        # if s3 enabled, download the map from s3
        file_path = self._map_config.file_path
        if self._s3config.enabled:
            client = S3Client(self._s3config)
            map_bytes = client.download(self._map_config.file_path)
            file_path = tempfile.mktemp()
            with open(file_path, "wb") as f:
                f.write(map_bytes)

        config_base64 = encode_to_base64(_generate_yaml_config(file_path))
        self._server_addr = f"localhost:{sim_port}"
        self._sim_proc = Popen(
            [
                self._sim_bin_path,
                "-config-data",
                config_base64,
                "-job",
                "agentsociety",
                "-listen",
                self._server_addr,
                "-syncer",
                "http://" + syncer_addr,
                "-cache",
                "",
                "-log.level",
                "warn",
            ],
            env=os.environ,
        )

        # step 1 tick as the syncer init
        await self.syncer.step()

        get_logger().info(
            f"start agentsociety-sim-oss at {self._server_addr}, PID={self._sim_proc.pid}"
        )

        # remove the temporary file
        if self._s3config.enabled:
            os.remove(file_path)

        super().init()

    async def close(self):
        """
        Terminate the simulation process if it's running.
        """
        if self._syncer is not None:
            await self._syncer.close()
            self._syncer = None
        if self._sim_proc is not None and self._sim_proc.poll() is None:
            get_logger().info(
                f"Terminating agentsociety-sim-oss at {self._server_addr}, PID={self._sim_proc.pid}, please ignore the PANIC message"
            )
            self._sim_proc.kill()
            # wait for the process to terminate
            self._sim_proc.wait()
        self._sim_proc = None

    async def step(self, n: int):
        assert n > 0, "`n` must >=1!"
        for _ in range(n):
            await self.syncer.step()
            self._tick += 1

    async def get_metrics(self) -> list[Tuple[str, float, int]]:
        """
        Get the metrics of the environment for recording.

        - **Returns**:
            - `List[Tuple[str, float, int]]`: A list of tuples, each containing the metric name, value, and step.
        """
        if self._last_metric_tick + self._environment_config.metric_interval > self._tick:
            return []

        # Add mobility metrics
        # 1. cumulative travel distance
        # 2. cumulative number of trips
        stat = await self._city_client.person_service.GetGlobalStatistics(
            {}, True
        )
        num_completed_trips = stat["num_completed_trips"]
        total_travel_time = stat["running_total_travel_time"]
        total_travel_distance = stat["running_total_travel_distance"]

        self._last_metric_tick = self._tick

        return [
            ("num_completed_trips", num_completed_trips, self._tick),
            ("total_travel_time (sec)", total_travel_time, self._tick),
            ("total_travel_distance (m)", total_travel_distance, self._tick),
        ]


def _generate_yaml_config(map_file) -> str:
    config_dict = {
        "input": {"map": {"file": os.path.abspath(map_file)}},
        "control": {
            "step": {"start": 0, "total": int(2**31 - 1), "interval": 1},
            "prefer_fixed_light": True,
        },
    }
    return yaml.dump(config_dict, allow_unicode=True)



from typing import Any, Awaitable, Callable, List, Optional, Union
import uuid

from pydantic import BaseModel, Field, field_serializer

from ..environment import MapConfig
from ..llm import LLMConfig
from .agent import AgentConfig, InstitutionAgentClass
from .env import EnvConfig
from .exp import (
    AgentFilterConfig,
    EnvironmentConfig,
    ExpConfig,
    WorkflowStepConfig,
    WorkflowType,
)
from ..taskloader import Task
from .utils import load_config_from_file

__all__ = [
    "EnvConfig",
    "AgentConfig",
    "WorkflowStepConfig",
    "ExpConfig",
    "EnvironmentConfig",
    "Config",
    "load_config_from_file",
    "InstitutionAgentClass",
    "WorkflowType",
    "AgentFilterConfig",
    "TaskLoaderConfig",
    "IndividualConfig",
]


class AgentsConfig(BaseModel):
    """Configuration for different types of agents in the simulation."""

    citizens: list[AgentConfig] = Field(..., min_length=1)
    """Citizen Agent configuration"""

    firms: list[AgentConfig] = Field(default=[])
    """Firm Agent configuration"""

    banks: list[AgentConfig] = Field(default=[])
    """Bank Agent configuration"""

    nbs: list[AgentConfig] = Field(default=[])
    """NBS Agent configuration"""

    governments: list[AgentConfig] = Field(default=[])
    """Government Agent configuration"""

    supervisor: Optional[AgentConfig] = Field(None)
    """Supervisor Agent configuration"""

    init_funcs: list[Callable[[Any], Union[None, Awaitable[None]]]] = Field([])
    """Initialization functions for simulation, the only one argument is the AgentSociety object"""

    @field_serializer("init_funcs")
    def serialize_init_funcs(self, init_funcs, info):
        return [func.__name__ for func in init_funcs]


class Config(BaseModel):
    """Configuration for the simulation."""

    llm: List[LLMConfig] = Field(..., min_length=1)
    """List of LLM configurations"""

    env: EnvConfig
    """Environment configuration"""

    map: MapConfig
    """Map configuration"""

    agents: AgentsConfig
    """Configuration for different types of agents in the simulation."""

    exp: ExpConfig
    """Experiment configuration"""

    logging_level: str = Field("INFO")
    """Logging level"""


class TaskLoaderConfig(BaseModel):
    """Configuration for the task loader."""

    task_type: type[Task]
    """Task type"""

    file_path: str
    """File path"""

    shuffle: bool = Field(default=False)
    """Shuffle the tasks"""


class IndividualConfig(BaseModel):
    """Configuration for the individual."""
    name: str = Field("default_individual")
    """Name of the solver"""

    llm: List[LLMConfig] = Field(..., min_length=1)
    """List of LLM configurations"""

    env: EnvConfig
    """Environment configuration"""

    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    """Unique identifier for the individual"""

    individual: AgentConfig
    """Individual Agent configuration"""

    task_loader: TaskLoaderConfig
    """Task loader configuration"""

    logging_level: str = Field("INFO")
    """Logging level"""

    @field_serializer("id")
    def serialize_id(self, id, info):
        return str(id)



import copy
from typing import Union, cast

from ..agent.distribution import Distribution, DistributionConfig
from ..cityagent.blocks.economy_block import EconomyBlock, EconomyBlockParams
from ..cityagent.blocks.mobility_block import MobilityBlock, MobilityBlockParams
from ..cityagent.blocks.other_block import OtherBlock, OtherBlockParams
from ..cityagent.blocks.social_block import SocialBlock, SocialBlockParams
from ..configs import InstitutionAgentClass, AgentConfig, Config
from .bankagent import BankAgent
from .firmagent import FirmAgent
from .governmentagent import GovernmentAgent
from .initial import bind_agent_info
from .memory_config import (
    DEFAULT_DISTRIBUTIONS,
    memory_config_bank,
    memory_config_firm,
    memory_config_government,
    memory_config_nbs,
    memory_config_societyagent,
)
from .nbsagent import NBSAgent
from .societyagent import SocietyAgent
from .sharing_params import (
    SocietyAgentConfig,
    SocietyAgentBlockOutput,
    SocietyAgentContext,
)

__all__ = [
    "default",
    "SocietyAgent",
    "FirmAgent",
    "BankAgent",
    "NBSAgent",
    "GovernmentAgent",
    "memory_config_societyagent",
    "memory_config_government",
    "memory_config_firm",
    "memory_config_bank",
    "memory_config_nbs",
    "SocietyAgentConfig",
    "SocietyAgentBlockOutput",
    "SocietyAgentContext",
]


BLOCK_MAPPING = {
    "MobilityBlock": MobilityBlock,
    "EconomyBlock": EconomyBlock,
    "SocialBlock": SocialBlock,
    "OtherBlock": OtherBlock,
}


def _fill_in_agent_class_and_memory_config(self: AgentConfig):
    if isinstance(self.agent_class, str):
        if self.agent_class == "citizen":
            self.agent_class = SocietyAgent
            if self.agent_params is not None:
                self.agent_params = SocietyAgent.ParamsType.model_validate(self.agent_params)
            if self.memory_config_func is None:
                self.memory_config_func = copy.deepcopy(memory_config_societyagent)
            distributions = cast(
                dict[str, Union[Distribution, DistributionConfig]],
                DEFAULT_DISTRIBUTIONS,
            )
            if self.memory_distributions is not None:
                distributions.update(self.memory_distributions)
            self.memory_distributions = copy.deepcopy(distributions)
            if self.blocks is None:
                self.blocks = {
                    MobilityBlock: MobilityBlockParams(),
                    EconomyBlock: EconomyBlockParams(),
                    SocialBlock: SocialBlockParams(),
                    OtherBlock: OtherBlockParams(),
                }
            else:
                blocks = {}
                for key, value in self.blocks.items():
                    if isinstance(key, str) and key in BLOCK_MAPPING:
                        blocks[BLOCK_MAPPING[key]] = BLOCK_MAPPING[key].ParamsType(
                            **value
                        )
                    else:
                        blocks[key] = value
                self.blocks = blocks
        elif self.agent_class == InstitutionAgentClass.FIRM.value:
            self.agent_class = FirmAgent
            if self.agent_params is not None:
                self.agent_params = FirmAgent.ParamsType(**self.agent_params)
            if self.memory_config_func is None:
                self.memory_config_func = memory_config_firm
        elif self.agent_class == InstitutionAgentClass.GOVERNMENT.value:
            self.agent_class = GovernmentAgent
            if self.agent_params is not None:
                self.agent_params = GovernmentAgent.ParamsType(**self.agent_params)
            if self.memory_config_func is None:
                self.memory_config_func = memory_config_government
        elif self.agent_class == InstitutionAgentClass.BANK.value:
            self.agent_class = BankAgent
            if self.agent_params is not None:
                self.agent_params = BankAgent.ParamsType(**self.agent_params)
            if self.memory_config_func is None:
                self.memory_config_func = memory_config_bank
        elif self.agent_class == InstitutionAgentClass.NBS.value:
            self.agent_class = NBSAgent
            if self.agent_params is not None:
                self.agent_params = NBSAgent.ParamsType(**self.agent_params)
            if self.memory_config_func is None:
                self.memory_config_func = memory_config_nbs
        else:
            pass
            # raise ValueError(f"Invalid agent class: {self.agent_class}")
    return self


def default(config: Config) -> Config:
    """
    Use the default values in cityagent to fill in the config.
    """
    # =====================
    # fill orgnizations
    # =====================
    if any(citizen_config.agent_class == "citizen" for citizen_config in config.agents.citizens):
        if len(config.agents.firms) == 0:
            config.agents.firms = [
                AgentConfig(
                    agent_class="firm",
                    number=1,
                )
            ]
        if len(config.agents.governments) == 0:
            config.agents.governments = [
                AgentConfig(
                    agent_class="government",
                    number=1,
                )
            ]
        if len(config.agents.banks) == 0:
            config.agents.banks = [
                AgentConfig(
                    agent_class="bank",
                    number=1,
                )
            ]
        if len(config.agents.nbs) == 0:
            config.agents.nbs = [
                AgentConfig(
                    agent_class="nbs",
                    number=1,
                )
            ]
    # =====================
    # agent config
    # =====================
    config.agents.citizens = [
        _fill_in_agent_class_and_memory_config(agent_config)
        for agent_config in config.agents.citizens
    ]
    config.agents.firms = [
        _fill_in_agent_class_and_memory_config(agent_config)
        for agent_config in config.agents.firms
    ]
    config.agents.governments = [
        _fill_in_agent_class_and_memory_config(agent_config)
        for agent_config in config.agents.governments
    ]
    config.agents.banks = [
        _fill_in_agent_class_and_memory_config(agent_config)
        for agent_config in config.agents.banks
    ]
    config.agents.nbs = [
        _fill_in_agent_class_and_memory_config(agent_config)
        for agent_config in config.agents.nbs
    ]
    if config.agents.supervisor is not None:
        config.agents.supervisor = _fill_in_agent_class_and_memory_config(
            config.agents.supervisor
        )
    # =====================
    # init functions
    # =====================
    if len(config.agents.init_funcs) == 0:
        config.agents.init_funcs = [
            bind_agent_info,
        ]
    return config



"""
A clear version of the simulation.
"""

from typing import Union


from ..configs import (
    Config,
    IndividualConfig,
)
from .simulationengine import SimulationEngine
from .individualengine import IndividualEngine

__all__ = ["AgentSociety"]


class AgentSociety:
    """
    Factory class for creating simulation engines based on configuration type.
    
    - **Description**:
        - A factory class that creates and returns the appropriate engine instance
        based on the configuration type provided.
        - Returns SimulationEngine for Config type
        - Returns TaskSolverEngine for SolverConfig type
    """
    
    @staticmethod
    def create(
        config: Union[Config, IndividualConfig],
        tenant_id: str = "",
    ) -> Union[SimulationEngine, IndividualEngine]:
        """
        Create and return the appropriate engine instance based on configuration type.
        
        - **Description**:
            - Factory method that creates the appropriate engine instance based on the
            configuration type provided.
            - For Config type: returns SimulationEngine instance
            - For IndividualConfig type: returns IndividualEngine instance
            
        - **Args**:
            - `config` (Union[Config, IndividualConfig]): The configuration object that determines
            which engine to create.
            - `tenant_id` (str, optional): The tenant ID for the engine. Defaults to "".
            
        - **Returns**:
            - `Union[SimulationEngine, IndividualEngine]`: The appropriate engine instance
            based on the configuration type.
            
        - **Raises**:
            - `ValueError`: If the configuration type is not supported.
        """
        if isinstance(config, Config):
            return SimulationEngine(config, tenant_id)
        elif isinstance(config, IndividualConfig):
            return IndividualEngine(config, tenant_id)
        else:
            raise ValueError(f"Invalid config type: {type(config)}. Expected Config or IndividualConfig.")

    def __init__(
        self,
        config: Union[Config, IndividualConfig],
        tenant_id: str = "",
    ) -> None:
        """
        Initialize AgentSociety factory class.
        
        - **Description**:
            - This constructor is kept for backward compatibility but is deprecated.
            - Use AgentSociety.create() static method instead.
            
        - **Args**:
            - `config` (Union[Config, IndividualConfig]): The configuration object.
            - `tenant_id` (str, optional): The tenant ID. Defaults to "".
        """
        import warnings
        warnings.warn(
            "AgentSociety constructor is deprecated. Use AgentSociety.create() instead.",
            DeprecationWarning,
            stacklevel=2
        )
        
        # For backward compatibility, delegate to the factory method
        engine = self.create(config, tenant_id)
        
        # Copy all attributes from the engine to this instance
        for attr_name in dir(engine):
            if not attr_name.startswith('_'):
                attr_value = getattr(engine, attr_name)
                if not callable(attr_value) or isinstance(attr_value, property):
                    setattr(self, attr_name, attr_value)
        
        # Store the engine instance for method delegation
        self._engine = engine

    def __getattr__(self, name):
        """
        Delegate attribute access to the underlying engine instance.
        
        - **Description**:
            - Delegates any attribute access to the underlying engine instance
            for backward compatibility.
            
        - **Args**:
            - `name` (str): The name of the attribute to access.
            
        - **Returns**:
            - The attribute value from the underlying engine.
        """
        if hasattr(self, '_engine'):
            return getattr(self._engine, name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")



from typing import Generic, List, TypeVar

from pydantic import BaseModel, Field

__all__ = ["ApiResponseWrapper"]

T = TypeVar("T")


class ApiResponseWrapper(BaseModel, Generic[T]):
    """API Response Wrapper"""

    data: T = Field(..., description="Response Data")


# 分页响应模型
class ApiPaginatedResponseWrapper(BaseModel, Generic[T]):
    """分页响应模型"""

    total: int = Field(..., description="总记录数")
    data: List[T] = Field(..., description="数据项列表")



from .cognition_block import CognitionBlock
from .economy_block import EconomyBlock
from .mobility_block import MobilityBlock
from .needs_block import NeedsBlock
from .other_block import OtherBlock
from .plan_block import PlanBlock
from .social_block import SocialBlock

__all__ = [
    "MobilityBlock",
    "CognitionBlock",
    "PlanBlock",
    "NeedsBlock",
    "SocialBlock",
    "EconomyBlock",
    "OtherBlock",
]



import re
from typing import Any, Optional

from openai.types.chat import ChatCompletionMessageParam
from ..memory import Memory


class FormatPrompt:
    """
    A class to handle the formatting of prompts based on a template,
    with support for system prompts and variable extraction.

    - **Attributes**:
        - `template` (str): The template string containing placeholders.
        - `system_prompt` (Optional[str]): An optional system prompt to add to the dialog.
        - `variables` (List[str]): A list of variable names extracted from the template.
        - `formatted_string` (str): The formatted string derived from the template and provided variables.
        - `bound_objects` (Dict[str, Any]): Dictionary of objects bound to the prompt for use in expressions.
    """

    def __init__(
        self,
        template: str,
        format_prompt: Optional[str] = None,
        system_prompt: Optional[str] = None,
        memory: Optional[Memory] = None,
    ) -> None:
        """
        - **Description**:
            - Initializes the FormatPrompt with a template, an optional system prompt, and bound objects.

        - **Args**:
            - `template` (str): The string template with variable placeholders.
            - `system_prompt` (Optional[str], optional): An optional system prompt. Defaults to None.
            - `**bound_objects`: Named objects to bind to the prompt for use in expressions.
        """
        self.template = template
        self.format_prompt = format_prompt  # Store the format prompt
        self.system_prompt = system_prompt  # Store the system prompt
        self.variables = self._extract_variables()
        self.formatted_string = ""  # To store the formatted string
        self.memory = memory  # Store memory

    def _extract_variables(self) -> list[str]:
        """
        - **Description**:
            - Extracts variable names from the template string using regular expressions.
            - Looks for simple variables in the format {variable_name}.

        - **Returns**:
            - `List[str]`: A list of variable names found within the template.
        """
        return re.findall(r"\{(\w+)\}", self.template)

    def _is_safe_expression(self, expr: str) -> bool:
        """
        - **Description**:
            - Checks if an expression is safe to evaluate.
            - Only allows attribute access and dictionary access.
            - Supports both single and double quotes for dictionary keys.
            - Supports numeric index access.

        - **Args**:
            - `expr` (str): The expression to check.

        - **Returns**:
            - `bool`: True if the expression is safe, False otherwise.
        """
        # Support both string keys and numeric indices
        safe_pattern = r'^(profile|status|context)(\.[a-zA-Z_][a-zA-Z0-9_]*|\[(?:[\'"][a-zA-Z0-9_][a-zA-Z0-9_]*[\'"]|\d+)\])*$'
        return bool(re.match(safe_pattern, expr))

    async def _eval_expr(self, expr: str, context: dict) -> Any:
        """
        - **Description**:
            - Evaluates expressions using eval with safety checks.
            - Supports expressions like profile.xxx, status.xxx, context.xxx
            - Also supports square bracket notation like profile.xxx["yyy"]
            - For profile and status, uses async get method from memory
            - Supports nested dictionary access with square brackets

        - **Args**:
            - `expr` (str): The expression to evaluate.
            - `context` (dict): The evaluation context.

        - **Returns**:
            - `Any`: The result of the expression evaluation.
        """
        # Create evaluation context
        eval_context = {
            "profile": self.memory.status if self.memory else {},
            "status": self.memory.status if self.memory else {},
            "context": context or {},
        }

        # Add safety check for the expression
        if not self._is_safe_expression(expr):
            raise ValueError(f"Unsafe expression: {expr}")

        try:
            # Parse the expression to handle profile and status differently
            if expr.startswith("context."):
                key = expr.split(".", 1)[1]
                return eval_context["context"].get(key, "Don't know")
            elif expr.startswith(("profile.", "status.")):
                # Get the base value using async get method
                base_key = expr.split(".", 1)[1].split("[")[0]
                base_value = (
                    await self.memory.status.get(base_key) if self.memory else None
                )

                # If there's no square bracket notation, return the base value
                if "[" not in expr:
                    return base_value

                # Extract the square bracket expression
                bracket_expr = expr[expr.find("[") :]
                # Create a safe evaluation context with the base value
                safe_context = {"value": base_value}
                # Evaluate the bracket expression
                return eval(f"value{bracket_expr}", {"__builtins__": {}}, safe_context)
            else:
                # For other expressions, use regular eval
                return eval(expr, {"__builtins__": {}}, eval_context)
        except Exception as e:
            print(f"Error evaluating expression '{expr}': {str(e)}")
            return "Don't know"

    async def format(
        self,
        context: Optional[dict] = None,
        **kwargs,
    ) -> str:
        """
        - **Description**:
            - Formats the template string using the provided context and keyword arguments.
            - Supports simple variables {var} for direct substitution.
            - Supports complex expressions ${expression} in three formats:
              - ${profile.xxx}: Retrieves values from memory.profile
              - ${status.xxx}: Retrieves values from memory.status
              - ${context.xxx}: Retrieves values from the provided context dictionary

        - **Args**:
            - `context` (Optional[dict]): Dictionary containing context values.
            - `**kwargs`: Variable names and their corresponding values to format the template.

        - **Returns**:
            - `str`: The formatted string.

        - **Raises**:
            - `KeyError`: If a placeholder in the template does not have a corresponding key in kwargs.
            - `ValueError`: If an expression has an invalid format.
        """
        # Create a dictionary to hold all formatting variables
        format_vars = {}

        # First add explicitly provided kwargs, these take highest precedence
        for key, value in kwargs.items():
            if key in self.variables:
                format_vars[key] = value

        eval_context = context if context else {}

        # Handle complex expressions in the template using ${expression} syntax
        complex_pattern = r"\$\{([^}]+?)\}"
        result = self.template

        # First, protect all ${...} expressions with a temporary marker
        protected_expressions = {}
        counter = 0

        def protect_expression(match):
            nonlocal counter
            placeholder = f"__EXPR_{counter}__"
            protected_expressions[placeholder] = match.group(0)
            counter += 1
            return placeholder

        # Replace all ${...} expressions with placeholders
        result = re.sub(complex_pattern, protect_expression, result)

        # Format simple variables
        try:
            result = result.format(**format_vars)
        except KeyError as e:
            raise KeyError(f"Missing required variable in template: {e}")
        except Exception as e:
            raise ValueError(f"Error formatting template: {e}")

        # Now evaluate and replace the protected expressions
        for placeholder, original_expr in protected_expressions.items():
            expr = original_expr[2:-1].strip()  # Remove ${ and }
            try:
                eval_result = await self._eval_expr(expr, eval_context)
                result = result.replace(
                    placeholder, str(eval_result) if eval_result is not None else ""
                )
            except Exception as e:
                print(f"Error evaluating expression '{expr}': {str(e)}")
                result = result.replace(placeholder, original_expr)

        self.formatted_string = result
        return self.formatted_string

    def to_dialog(self) -> list[ChatCompletionMessageParam]:
        """
        - **Description**:
            - Converts the formatted prompt and optional system prompt into a dialog format suitable for chat systems.

        - **Returns**:
            - `List[Dict[str, str]]`: A list representing the dialog with roles and content.
        """
        dialog = []
        if self.system_prompt:
            dialog.append(
                {"role": "system", "content": self.system_prompt}
            )  # Add system prompt if it exists
        if self.format_prompt:
            dialog.append(
                {
                    "role": "user",
                    "content": self.formatted_string + "\n" + self.format_prompt,
                }
            )
        else:
            dialog.append(
                {"role": "user", "content": self.formatted_string}
            )  # Add user content
        return dialog

    def log(self) -> None:
        """
        - **Description**:
            - Logs the details of the FormatPrompt instance, including the template,
              system prompt, extracted variables, and formatted string.
        """
        print(f"Prompt Template: {self.template}")
        print(f"Format Prompt: {self.format_prompt}")
        print(f"System Prompt: {self.system_prompt}")  # Log the system prompt
        print(f"Variables: {self.variables}")
        print(f"Formatted String: {self.formatted_string}")  # Log the formatted string



import random
from collections import deque
from typing import Optional

import numpy as np

from ..agent.distribution import (
    ChoiceDistribution,
    Distribution,
    UniformIntDistribution,
    sample_field_value,
)
from ..agent.memory_config_generator import MemoryConfig, MemoryAttribute
from ..environment.economy import EconomyEntityType

pareto_param = 8
payment_max_skill_multiplier_base = 950
payment_max_skill_multiplier = float(payment_max_skill_multiplier_base)
pmsm = payment_max_skill_multiplier
pareto_samples = np.random.pareto(pareto_param, size=(1000, 10))
clipped_skills = np.minimum(pmsm, (pmsm - 1) * pareto_samples + 1)
sorted_clipped_skills = np.sort(clipped_skills, axis=1)
agent_skills = list(sorted_clipped_skills.mean(axis=0))

__all__ = [
    "memory_config_societyagent",
    "memory_config_firm",
    "memory_config_government",
    "memory_config_bank",
    "memory_config_nbs",
    "DEFAULT_DISTRIBUTIONS",
]


# Default distributions for different profile fields
DEFAULT_DISTRIBUTIONS: dict[str, Distribution] = {
    "name": ChoiceDistribution(
        choices=[
            "Alice",
            "Bob",
            "Charlie",
            "David",
            "Eve",
            "Frank",
            "Grace",
            "Helen",
            "Ivy",
            "Jack",
            "Kelly",
            "Lily",
            "Mike",
            "Nancy",
            "Oscar",
            "Peter",
            "Queen",
            "Rose",
            "Sam",
            "Tom",
            "Ulysses",
            "Vicky",
            "Will",
            "Xavier",
            "Yvonne",
            "Zack",
        ]
    ),
    "gender": ChoiceDistribution(choices=["Male", "Female"]),
    "age": UniformIntDistribution(min_value=18, max_value=65),
    "education": ChoiceDistribution(
        choices=["Doctor", "Master", "Bachelor", "College", "High School"]
    ),
    "occupation": ChoiceDistribution(
        choices=[
            "Student",
            "Teacher",
            "Doctor",
            "Engineer",
            "Manager",
            "Businessman",
            "Artist",
            "Athlete",
            "Other",
        ]
    ),
    "marriage_status": ChoiceDistribution(
        choices=["Not married", "Married", "Divorced", "Widowed"]
    ),
}


def memory_config_societyagent(
    distributions: dict[str, Distribution],
    class_config: Optional[list[MemoryAttribute]] = None,
) -> MemoryConfig:
    """Generate memory configuration for society agents."""
    attributes = {
        "type": MemoryAttribute(
            name="type",
            type=str,
            default_or_value="citizen",
            description="agent type",
            whether_embedding=False,
        ),
        # Needs Model
        "hunger_satisfaction": MemoryAttribute(
            name="hunger_satisfaction",
            type=float,
            default_or_value=0.9,
            description="hunger satisfaction",
            whether_embedding=False,
        ),
        "energy_satisfaction": MemoryAttribute(
            name="energy_satisfaction",
            type=float,
            default_or_value=0.9,
            description="energy satisfaction",
            whether_embedding=False,
        ),
        "safety_satisfaction": MemoryAttribute(
            name="safety_satisfaction",
            type=float,
            default_or_value=0.4,
            description="safety satisfaction",
            whether_embedding=False,
        ),
        "social_satisfaction": MemoryAttribute(
            name="social_satisfaction",
            type=float,
            default_or_value=0.6,
            description="social satisfaction",
            whether_embedding=False,
        ),
        "current_need": MemoryAttribute(
            name="current_need",
            type=str,
            default_or_value="none",
            description="current need",
            whether_embedding=False,
        ),
        # Plan Behavior Model
        "current_plan": MemoryAttribute(
            name="current_plan",
            type=dict,
            default_or_value={},
            description="current plan",
            whether_embedding=False,
        ),
        "execution_context": MemoryAttribute(
            name="execution_context",
            type=dict,
            default_or_value={},
            description="execution context",
            whether_embedding=False,
        ),
        "plan_history": MemoryAttribute(
            name="plan_history",
            type=list,
            default_or_value=[],
            description="plan history",
            whether_embedding=False,
        ),
        # cognition
        "emotion": MemoryAttribute(
            name="emotion",
            type=dict,
            default_or_value={
                "sadness": 5,
                "joy": 5,
                "fear": 5,
                "disgust": 5,
                "anger": 5,
                "surprise": 5,
            },
            description="emotion state",
            whether_embedding=False,
        ),
        "attitude": MemoryAttribute(
            name="attitude",
            type=dict,
            default_or_value={},
            description="attitude",
            whether_embedding=True,
        ),
        "thought": MemoryAttribute(
            name="thought",
            type=str,
            default_or_value="Currently nothing good or bad is happening",
            description="current thought",
            whether_embedding=True,
        ),
        "emotion_types": MemoryAttribute(
            name="emotion_types",
            type=str,
            default_or_value="Relief",
            description="emotion types",
            whether_embedding=True,
        ),
        # economy
        "work_skill": MemoryAttribute(
            name="work_skill",
            type=float,
            default_or_value=random.choice(agent_skills),
            description="work skill",
            whether_embedding=True,
        ),
        "tax_paid": MemoryAttribute(
            name="tax_paid",
            type=float,
            default_or_value=0.0,
            description="tax paid",
            whether_embedding=False,
        ),
        "consumption_currency": MemoryAttribute(
            name="consumption_currency",
            type=float,
            default_or_value=0.0,
            description="consumption currency",
            whether_embedding=False,
        ),
        "goods_demand": MemoryAttribute(
            name="goods_demand",
            type=int,
            default_or_value=0,
            description="goods demand",
            whether_embedding=False,
        ),
        "goods_consumption": MemoryAttribute(
            name="goods_consumption",
            type=int,
            default_or_value=0,
            description="goods consumption",
            whether_embedding=False,
        ),
        "work_propensity": MemoryAttribute(
            name="work_propensity",
            type=float,
            default_or_value=0.0,
            description="work propensity",
            whether_embedding=False,
        ),
        "consumption_propensity": MemoryAttribute(
            name="consumption_propensity",
            type=float,
            default_or_value=0.0,
            description="consumption propensity",
            whether_embedding=False,
        ),
        "to_consumption_currency": MemoryAttribute(
            name="to_consumption_currency",
            type=float,
            default_or_value=0.0,
            description="to consumption currency",
            whether_embedding=False,
        ),
        "firm_id": MemoryAttribute(
            name="firm_id",
            type=int,
            default_or_value=0,
            description="firm id",
            whether_embedding=False,
        ),
        "government_id": MemoryAttribute(
            name="government_id",
            type=int,
            default_or_value=0,
            description="government id",
            whether_embedding=False,
        ),
        "bank_id": MemoryAttribute(
            name="bank_id",
            type=int,
            default_or_value=0,
            description="bank id",
            whether_embedding=False,
        ),
        "nbs_id": MemoryAttribute(
            name="nbs_id",
            type=int,
            default_or_value=0,
            description="nbs id",
            whether_embedding=False,
        ),
        "dialog_queue": MemoryAttribute(
            name="dialog_queue",
            type=deque,
            default_or_value=deque(maxlen=3),
            description="dialog queue",
            whether_embedding=False,
        ),
        "firm_forward": MemoryAttribute(
            name="firm_forward",
            type=int,
            default_or_value=0,
            description="firm forward",
            whether_embedding=False,
        ),
        "bank_forward": MemoryAttribute(
            name="bank_forward",
            type=int,
            default_or_value=0,
            description="bank forward",
            whether_embedding=False,
        ),
        "nbs_forward": MemoryAttribute(
            name="nbs_forward",
            type=int,
            default_or_value=0,
            description="nbs forward",
            whether_embedding=False,
        ),
        "government_forward": MemoryAttribute(
            name="government_forward",
            type=int,
            default_or_value=0,
            description="government forward",
            whether_embedding=False,
        ),
        "forward": MemoryAttribute(
            name="forward",
            type=int,
            default_or_value=0,
            description="forward",
            whether_embedding=False,
        ),
        "depression": MemoryAttribute(
            name="depression",
            type=float,
            default_or_value=0.0,
            description="depression",
            whether_embedding=False,
        ),
        "ubi_opinion": MemoryAttribute(
            name="ubi_opinion",
            type=list,
            default_or_value=[],
            description="ubi opinion",
            whether_embedding=False,
        ),
        "working_experience": MemoryAttribute(
            name="working_experience",
            type=list,
            default_or_value=[],
            description="working experience",
            whether_embedding=False,
        ),
        "work_hour_month": MemoryAttribute(
            name="work_hour_month",
            type=float,
            default_or_value=160,
            description="work hour month",
            whether_embedding=False,
        ),
        "work_hour_finish": MemoryAttribute(
            name="work_hour_finish",
            type=float,
            default_or_value=0,
            description="work hour finish",
            whether_embedding=False,
        ),
        # social
        "friends": MemoryAttribute(
            name="friends",
            type=list,
            default_or_value=[],
            description="friends list",
            whether_embedding=False,
        ),
        "public_friends": MemoryAttribute(
            name="public_friends",
            type=list,
            default_or_value=[],
            description="public friends list",
            whether_embedding=False,
        ),
        "relationships": MemoryAttribute(
            name="relationships",
            type=dict,
            default_or_value={},
            description="relationship strength with each friend",
            whether_embedding=False,
        ),
        "relation_types": MemoryAttribute(
            name="relation_types",
            type=dict,
            default_or_value={},
            description="relation types",
            whether_embedding=False,
        ),
        "chat_histories": MemoryAttribute(
            name="chat_histories",
            type=dict,
            default_or_value={},
            description="all chat histories",
            whether_embedding=False,
        ),
        "interactions": MemoryAttribute(
            name="interactions",
            type=dict,
            default_or_value={},
            description="all interaction records",
            whether_embedding=False,
        ),
        # mobility
        "number_poi_visited": MemoryAttribute(
            name="number_poi_visited",
            type=int,
            default_or_value=1,
            description="number of poi visited",
            whether_embedding=False,
        ),
        "location_knowledge": MemoryAttribute(
            name="location_knowledge",
            type=dict,
            default_or_value={},
            description="location knowledge",
            whether_embedding=False,
        ),
    }

    # Add profile attributes
    profile_attributes = {
        "name": MemoryAttribute(
            name="name",
            type=str,
            default_or_value=sample_field_value(distributions, "name"),
            description="agent's name",
            whether_embedding=True,
        ),
        "gender": MemoryAttribute(
            name="gender",
            type=str,
            default_or_value=sample_field_value(distributions, "gender"),
            description="agent's gender",
            whether_embedding=True,
        ),
        "age": MemoryAttribute(
            name="age",
            type=int,
            default_or_value=sample_field_value(distributions, "age"),
            description="agent's age group",
            whether_embedding=True,
        ),
        "education": MemoryAttribute(
            name="education",
            type=str,
            default_or_value=sample_field_value(distributions, "education"),
            description="agent's education level",
            whether_embedding=True,
        ),
        "skill": MemoryAttribute(
            name="skill",
            type=str,
            default_or_value="unknown",
            description="agent's skills",
            whether_embedding=True,
        ),
        "occupation": MemoryAttribute(
            name="occupation",
            type=str,
            default_or_value=sample_field_value(distributions, "occupation"),
            description="agent's occupation",
            whether_embedding=True,
        ),
        "family_consumption": MemoryAttribute(
            name="family_consumption",
            type=str,
            default_or_value="unknown",
            description="agent's family consumption pattern",
            whether_embedding=True,
        ),
        "consumption": MemoryAttribute(
            name="consumption",
            type=str,
            default_or_value="unknown",
            description="agent's consumption pattern",
            whether_embedding=True,
        ),
        "personality": MemoryAttribute(
            name="personality",
            type=str,
            default_or_value="unknown",
            description="agent's personality",
            whether_embedding=True,
        ),
        "income": MemoryAttribute(
            name="income",
            type=float,
            default_or_value=5000,
            description="agent's income",
            whether_embedding=True,
        ),
        "currency": MemoryAttribute(
            name="currency",
            type=float,
            default_or_value=30000,
            description="agent's currency",
            whether_embedding=True,
        ),
        "residence": MemoryAttribute(
            name="residence",
            type=str,
            default_or_value="unknown",
            description="agent's residence",
            whether_embedding=True,
        ),
        "city": MemoryAttribute(
            name="city",
            type=str,
            default_or_value="unknown",
            description="agent's city",
            whether_embedding=True,
        ),
        "race": MemoryAttribute(
            name="race",
            type=str,
            default_or_value="unknown",
            description="agent's race",
            whether_embedding=True,
        ),
        "religion": MemoryAttribute(
            name="religion",
            type=str,
            default_or_value="unknown",
            description="agent's religion",
            whether_embedding=True,
        ),
        "marriage_status": MemoryAttribute(
            name="marriage_status",
            type=str,
            default_or_value=sample_field_value(distributions, "marriage_status"),
            description="agent's marriage status",
            whether_embedding=True,
        ),
        "background_story": MemoryAttribute(
            name="background_story",
            type=str,
            default_or_value="No background story",
            description="agent's background story",
            whether_embedding=True,
        ),
    }

    # Add base attributes
    base_attributes = {
        "home": MemoryAttribute(
            name="home",
            type=dict,
            default_or_value={
                "aoi_position": {
                    "aoi_id": sample_field_value(distributions, "home_aoi_id")
                }
            },
            description="agent's home location",
            whether_embedding=False,
        ),
        "work": MemoryAttribute(
            name="work",
            type=dict,
            default_or_value={
                "aoi_position": {
                    "aoi_id": sample_field_value(distributions, "work_aoi_id")
                }
            },
            description="agent's work location",
            whether_embedding=False,
        ),
    }

    # Merge all attributes
    all_attributes = {**attributes, **profile_attributes, **base_attributes}

    # Add class-specific attributes if provided
    if class_config:
        for attr in class_config:
            if attr.name in all_attributes:
                continue
            all_attributes[attr.name] = attr

    return MemoryConfig(attributes=all_attributes)


def memory_config_firm(
    distributions: dict[str, Distribution],
    class_config: Optional[list[MemoryAttribute]] = None,
) -> MemoryConfig:
    """Generate memory configuration for firm agents."""
    attributes = {
        "type": MemoryAttribute(
            name="type",
            type=int,
            default_or_value=EconomyEntityType.Firm,
            description="agent type",
            whether_embedding=False,
        ),
        "location": MemoryAttribute(
            name="location",
            type=dict,
            default_or_value={
                "aoi_position": {
                    "aoi_id": sample_field_value(distributions, "aoi_id")
                }
            },
            description="firm location",
            whether_embedding=False,
        ),
        "price": MemoryAttribute(
            name="price",
            type=float,
            default_or_value=float(np.mean(agent_skills)),
            description="firm price",
            whether_embedding=False,
        ),
        "inventory": MemoryAttribute(
            name="inventory",
            type=int,
            default_or_value=0,
            description="firm inventory",
            whether_embedding=False,
        ),
        "employees": MemoryAttribute(
            name="employees",
            type=list,
            default_or_value=[],
            description="firm employees",
            whether_embedding=False,
        ),
        "employees_agent_id": MemoryAttribute(
            name="employees_agent_id",
            type=list,
            default_or_value=[],
            description="firm employees agent id",
            whether_embedding=False,
        ),
        "nominal_gdp": MemoryAttribute(
            name="nominal_gdp",
            type=list,
            default_or_value=[],
            description="nominal gdp",
            whether_embedding=False,
        ),
        "real_gdp": MemoryAttribute(
            name="real_gdp",
            type=list,
            default_or_value=[],
            description="real gdp",
            whether_embedding=False,
        ),
        "unemployment": MemoryAttribute(
            name="unemployment",
            type=list,
            default_or_value=[],
            description="unemployment",
            whether_embedding=False,
        ),
        "wages": MemoryAttribute(
            name="wages",
            type=list,
            default_or_value=[],
            description="wages",
            whether_embedding=False,
        ),
        "demand": MemoryAttribute(
            name="demand",
            type=int,
            default_or_value=0,
            description="demand",
            whether_embedding=False,
        ),
        "sales": MemoryAttribute(
            name="sales",
            type=int,
            default_or_value=0,
            description="sales",
            whether_embedding=False,
        ),
        "prices": MemoryAttribute(
            name="prices",
            type=list,
            default_or_value=[float(np.mean(agent_skills))],
            description="prices",
            whether_embedding=False,
        ),
        "working_hours": MemoryAttribute(
            name="working_hours",
            type=list,
            default_or_value=[],
            description="working hours",
            whether_embedding=False,
        ),
        "depression": MemoryAttribute(
            name="depression",
            type=list,
            default_or_value=[],
            description="depression",
            whether_embedding=False,
        ),
        "consumption_currency": MemoryAttribute(
            name="consumption_currency",
            type=list,
            default_or_value=[],
            description="consumption currency",
            whether_embedding=False,
        ),
        "income_currency": MemoryAttribute(
            name="income_currency",
            type=list,
            default_or_value=[],
            description="income currency",
            whether_embedding=False,
        ),
        "locus_control": MemoryAttribute(
            name="locus_control",
            type=list,
            default_or_value=[],
            description="locus control",
            whether_embedding=False,
        ),
        "bracket_cutoffs": MemoryAttribute(
            name="bracket_cutoffs",
            type=list,
            default_or_value=list(np.array([0, 9875, 40125, 85525, 163300, 207350, 518400]) / 12),
            description="bracket cutoffs",
            whether_embedding=False,
        ),
        "bracket_rates": MemoryAttribute(
            name="bracket_rates",
            type=list,
            default_or_value=[0.1, 0.12, 0.22, 0.24, 0.32, 0.35, 0.37],
            description="bracket rates",
            whether_embedding=False,
        ),
        "interest_rate": MemoryAttribute(
            name="interest_rate",
            type=float,
            default_or_value=0.03,
            description="interest rate",
            whether_embedding=False,
        ),
        "citizen_ids": MemoryAttribute(
            name="citizen_ids",
            type=list,
            default_or_value=[],
            description="citizen ids",
            whether_embedding=False,
        ),
        "firm_id": MemoryAttribute(
            name="firm_id",
            type=int,
            default_or_value=0,
            description="firm id",
            whether_embedding=False,
        ),
        "currency": MemoryAttribute(
            name="currency",
            type=float,
            default_or_value=1e12,
            description="firm currency",
            whether_embedding=False,
        ),
    }

    # Add class-specific attributes if provided
    if class_config:
        for attr in class_config:
            if attr.name in attributes:
                continue
            attributes[attr.name] = attr

    return MemoryConfig(attributes=attributes)


def memory_config_government(
    distributions: dict[str, Distribution],
    class_config: Optional[list[MemoryAttribute]] = None,
) -> MemoryConfig:
    """Generate memory configuration for government agents."""
    attributes = {
        "type": MemoryAttribute(
            name="type",
            type=int,
            default_or_value=EconomyEntityType.Government,
            description="agent type",
            whether_embedding=False,
        ),
        "bracket_cutoffs": MemoryAttribute(
            name="bracket_cutoffs",
            type=list,
            default_or_value=list(np.array([0, 9875, 40125, 85525, 163300, 207350, 518400]) / 12),
            description="bracket cutoffs",
            whether_embedding=False,
        ),
        "bracket_rates": MemoryAttribute(
            name="bracket_rates",
            type=list,
            default_or_value=[0.1, 0.12, 0.22, 0.24, 0.32, 0.35, 0.37],
            description="bracket rates",
            whether_embedding=False,
        ),
        "citizen_ids": MemoryAttribute(
            name="citizen_ids",
            type=list,
            default_or_value=[],
            description="citizen ids",
            whether_embedding=False,
        ),
        "citizens_agent_id": MemoryAttribute(
            name="citizens_agent_id",
            type=list,
            default_or_value=[],
            description="citizens agent id",
            whether_embedding=False,
        ),
        "nominal_gdp": MemoryAttribute(
            name="nominal_gdp",
            type=list,
            default_or_value=[],
            description="nominal gdp",
            whether_embedding=False,
        ),
        "real_gdp": MemoryAttribute(
            name="real_gdp",
            type=list,
            default_or_value=[],
            description="real gdp",
            whether_embedding=False,
        ),
        "unemployment": MemoryAttribute(
            name="unemployment",
            type=list,
            default_or_value=[],
            description="unemployment",
            whether_embedding=False,
        ),
        "wages": MemoryAttribute(
            name="wages",
            type=list,
            default_or_value=[],
            description="wages",
            whether_embedding=False,
        ),
        "prices": MemoryAttribute(
            name="prices",
            type=list,
            default_or_value=[float(np.mean(agent_skills))],
            description="prices",
            whether_embedding=False,
        ),
        "working_hours": MemoryAttribute(
            name="working_hours",
            type=list,
            default_or_value=[],
            description="working hours",
            whether_embedding=False,
        ),
        "depression": MemoryAttribute(
            name="depression",
            type=list,
            default_or_value=[],
            description="depression",
            whether_embedding=False,
        ),
        "consumption_currency": MemoryAttribute(
            name="consumption_currency",
            type=list,
            default_or_value=[],
            description="consumption currency",
            whether_embedding=False,
        ),
        "income_currency": MemoryAttribute(
            name="income_currency",
            type=list,
            default_or_value=[],
            description="income currency",
            whether_embedding=False,
        ),
        "locus_control": MemoryAttribute(
            name="locus_control",
            type=list,
            default_or_value=[],
            description="locus control",
            whether_embedding=False,
        ),
        "inventory": MemoryAttribute(
            name="inventory",
            type=int,
            default_or_value=0,
            description="inventory",
            whether_embedding=False,
        ),
        "interest_rate": MemoryAttribute(
            name="interest_rate",
            type=float,
            default_or_value=0.03,
            description="interest rate",
            whether_embedding=False,
        ),
        "price": MemoryAttribute(
            name="price",
            type=float,
            default_or_value=float(np.mean(agent_skills)),
            description="price",
            whether_embedding=False,
        ),
        "employees": MemoryAttribute(
            name="employees",
            type=list,
            default_or_value=[],
            description="employees",
            whether_embedding=False,
        ),
        "firm_id": MemoryAttribute(
            name="firm_id",
            type=int,
            default_or_value=0,
            description="firm id",
            whether_embedding=False,
        ),
        "currency": MemoryAttribute(
            name="currency",
            type=float,
            default_or_value=1e12,
            description="government currency",
            whether_embedding=False,
        ),
    }

    # Add class-specific attributes if provided
    if class_config:
        for attr in class_config:
            if attr.name in attributes:
                continue
            attributes[attr.name] = attr

    return MemoryConfig(attributes=attributes)


def memory_config_bank(
    distributions: dict[str, Distribution],
    class_config: Optional[list[MemoryAttribute]] = None,
) -> MemoryConfig:
    """Generate memory configuration for bank agents."""
    attributes = {
        "type": MemoryAttribute(
            name="type",
            type=int,
            default_or_value=EconomyEntityType.Bank,
            description="agent type",
            whether_embedding=False,
        ),
        "interest_rate": MemoryAttribute(
            name="interest_rate",
            type=float,
            default_or_value=0.03,
            description="interest rate",
            whether_embedding=False,
        ),
        "citizen_ids": MemoryAttribute(
            name="citizen_ids",
            type=list,
            default_or_value=[],
            description="citizen ids",
            whether_embedding=False,
        ),
        "bracket_cutoffs": MemoryAttribute(
            name="bracket_cutoffs",
            type=list,
            default_or_value=list(np.array([0, 9875, 40125, 85525, 163300, 207350, 518400]) / 12),
            description="bracket cutoffs",
            whether_embedding=False,
        ),
        "bracket_rates": MemoryAttribute(
            name="bracket_rates",
            type=list,
            default_or_value=[0.1, 0.12, 0.22, 0.24, 0.32, 0.35, 0.37],
            description="bracket rates",
            whether_embedding=False,
        ),
        "inventory": MemoryAttribute(
            name="inventory",
            type=int,
            default_or_value=0,
            description="inventory",
            whether_embedding=False,
        ),
        "nominal_gdp": MemoryAttribute(
            name="nominal_gdp",
            type=list,
            default_or_value=[],
            description="nominal gdp",
            whether_embedding=False,
        ),
        "real_gdp": MemoryAttribute(
            name="real_gdp",
            type=list,
            default_or_value=[],
            description="real gdp",
            whether_embedding=False,
        ),
        "unemployment": MemoryAttribute(
            name="unemployment",
            type=list,
            default_or_value=[],
            description="unemployment",
            whether_embedding=False,
        ),
        "wages": MemoryAttribute(
            name="wages",
            type=list,
            default_or_value=[],
            description="wages",
            whether_embedding=False,
        ),
        "prices": MemoryAttribute(
            name="prices",
            type=list,
            default_or_value=[float(np.mean(agent_skills))],
            description="prices",
            whether_embedding=False,
        ),
        "working_hours": MemoryAttribute(
            name="working_hours",
            type=list,
            default_or_value=[],
            description="working hours",
            whether_embedding=False,
        ),
        "depression": MemoryAttribute(
            name="depression",
            type=list,
            default_or_value=[],
            description="depression",
            whether_embedding=False,
        ),
        "consumption_currency": MemoryAttribute(
            name="consumption_currency",
            type=list,
            default_or_value=[],
            description="consumption currency",
            whether_embedding=False,
        ),
        "income_currency": MemoryAttribute(
            name="income_currency",
            type=list,
            default_or_value=[],
            description="income currency",
            whether_embedding=False,
        ),
        "locus_control": MemoryAttribute(
            name="locus_control",
            type=list,
            default_or_value=[],
            description="locus control",
            whether_embedding=False,
        ),
        "price": MemoryAttribute(
            name="price",
            type=float,
            default_or_value=float(np.mean(agent_skills)),
            description="price",
            whether_embedding=False,
        ),
        "employees": MemoryAttribute(
            name="employees",
            type=list,
            default_or_value=[],
            description="employees",
            whether_embedding=False,
        ),
        "firm_id": MemoryAttribute(
            name="firm_id",
            type=int,
            default_or_value=0,
            description="firm id",
            whether_embedding=False,
        ),
        "currency": MemoryAttribute(
            name="currency",
            type=float,
            default_or_value=1e12,
            description="bank currency",
            whether_embedding=False,
        ),
    }

    # Add class-specific attributes if provided
    if class_config:
        for attr in class_config:
            if attr.name in attributes:
                continue
            attributes[attr.name] = attr

    return MemoryConfig(attributes=attributes)


def memory_config_nbs(
    distributions: dict[str, Distribution],
    class_config: Optional[list[MemoryAttribute]] = None,
) -> MemoryConfig:
    """Generate memory configuration for NBS agents."""
    attributes = {
        "type": MemoryAttribute(
            name="type",
            type=int,
            default_or_value=EconomyEntityType.NBS,
            description="agent type",
            whether_embedding=False,
        ),
        "citizen_ids": MemoryAttribute(
            name="citizen_ids",
            type=list,
            default_or_value=[],
            description="citizen ids",
            whether_embedding=False,
        ),
        "nominal_gdp": MemoryAttribute(
            name="nominal_gdp",
            type=dict,
            default_or_value={},
            description="nominal gdp",
            whether_embedding=False,
        ),
        "real_gdp": MemoryAttribute(
            name="real_gdp",
            type=dict,
            default_or_value={},
            description="real gdp",
            whether_embedding=False,
        ),
        "real_gdp_metric": MemoryAttribute(
            name="real_gdp_metric",
            type=float,
            default_or_value=0.0,
            description="real gdp metric",
            whether_embedding=False,
        ),
        "unemployment": MemoryAttribute(
            name="unemployment",
            type=dict,
            default_or_value={},
            description="unemployment",
            whether_embedding=False,
        ),
        "wages": MemoryAttribute(
            name="wages",
            type=dict,
            default_or_value={},
            description="wages",
            whether_embedding=False,
        ),
        "prices": MemoryAttribute(
            name="prices",
            type=dict,
            default_or_value={"0": float(np.mean(agent_skills))},
            description="prices",
            whether_embedding=False,
        ),
        "working_hours": MemoryAttribute(
            name="working_hours",
            type=dict,
            default_or_value={},
            description="working hours",
            whether_embedding=False,
        ),
        "working_hours_metric": MemoryAttribute(
            name="working_hours_metric",
            type=float,
            default_or_value=0.0,
            description="working hours metric",
            whether_embedding=False,
        ),
        "depression": MemoryAttribute(
            name="depression",
            type=dict,
            default_or_value={},
            description="depression",
            whether_embedding=False,
        ),
        "depression_metric": MemoryAttribute(
            name="depression_metric",
            type=float,
            default_or_value=0.0,
            description="depression metric",
            whether_embedding=False,
        ),
        "consumption_currency": MemoryAttribute(
            name="consumption_currency",
            type=dict,
            default_or_value={},
            description="consumption currency",
            whether_embedding=False,
        ),
        "consumption_currency_metric": MemoryAttribute(
            name="consumption_currency_metric",
            type=float,
            default_or_value=0.0,
            description="consumption currency metric",
            whether_embedding=False,
        ),
        "income_currency": MemoryAttribute(
            name="income_currency",
            type=dict,
            default_or_value={},
            description="income currency",
            whether_embedding=False,
        ),
        "income_currency_metric": MemoryAttribute(
            name="income_currency_metric",
            type=float,
            default_or_value=0.0,
            description="income currency metric",
            whether_embedding=False,
        ),
        "locus_control": MemoryAttribute(
            name="locus_control",
            type=dict,
            default_or_value={},
            description="locus control",
            whether_embedding=False,
        ),
        "firm_id": MemoryAttribute(
            name="firm_id",
            type=int,
            default_or_value=0,
            description="firm id",
            whether_embedding=False,
        ),
        "bracket_cutoffs": MemoryAttribute(
            name="bracket_cutoffs",
            type=list,
            default_or_value=list(np.array([0, 9875, 40125, 85525, 163300, 207350, 518400]) / 12),
            description="bracket cutoffs",
            whether_embedding=False,
        ),
        "bracket_rates": MemoryAttribute(
            name="bracket_rates",
            type=list,
            default_or_value=[0.1, 0.12, 0.22, 0.24, 0.32, 0.35, 0.37],
            description="bracket rates",
            whether_embedding=False,
        ),
        "inventory": MemoryAttribute(
            name="inventory",
            type=int,
            default_or_value=0,
            description="inventory",
            whether_embedding=False,
        ),
        "interest_rate": MemoryAttribute(
            name="interest_rate",
            type=float,
            default_or_value=0.03,
            description="interest rate",
            whether_embedding=False,
        ),
        "price": MemoryAttribute(
            name="price",
            type=float,
            default_or_value=float(np.mean(agent_skills)),
            description="price",
            whether_embedding=False,
        ),
        "price_metric": MemoryAttribute(
            name="price_metric",
            type=float,
            default_or_value=0.0,
            description="price metric",
            whether_embedding=False,
        ),
        "employees": MemoryAttribute(
            name="employees",
            type=list,
            default_or_value=[],
            description="employees",
            whether_embedding=False,
        ),
        "forward_times": MemoryAttribute(
            name="forward_times",
            type=int,
            default_or_value=0,
            description="forward times",
            whether_embedding=False,
        ),
        "currency": MemoryAttribute(
            name="currency",
            type=float,
            default_or_value=1e12,
            description="nbs currency",
            whether_embedding=False,
        ),
    }

    # Add class-specific attributes if provided
    if class_config:
        for attr in class_config:
            if attr.name in attributes:
                continue
            attributes[attr.name] = attr

    return MemoryConfig(attributes=attributes)
